"""
Module for Serialization and Deserialization of KNX Session Authenticates.

The SESSION_AUTHENTICATE shall be sent by the KNXnet/IP secure client to the
control endpoint of the KNXnet/IP secure server after the Diffie-Hellman handshake
to authenticate the user against the server device.
"""

from __future__ import annotations

from typing import Final

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .knxip_enum import KNXIPServiceType


class SessionAuthenticate(KNXIPBody):
    """Representation of a KNX Session Authenticate."""

    SERVICE_TYPE = KNXIPServiceType.SESSION_AUTHENTICATE
    LENGTH: Final = 18

    def __init__(
        self,
        user_id: int = 0x02,
        message_authentication_code: bytes = bytes(16),
    ) -> None:
        """Initialize SessionAuthenticate object."""
        # 00h: Reserved, shall not be used
        # 01h: Management level access
        # 02h - 7Fh: User level access
        # 80h - FFh: Reserved, shall not be used
        # TODO: maybe use an Enum instead of int or raise an error in
        #   to_knx() and handle in RequestResponse class
        self.user_id = user_id
        self.message_authentication_code = message_authentication_code

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return SessionAuthenticate.LENGTH

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) != SessionAuthenticate.LENGTH:
            raise CouldNotParseKNXIP("SessionAuthenticate has wrong length")
        self.user_id = raw[1]
        self.message_authentication_code = raw[2:]
        return SessionAuthenticate.LENGTH

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            bytes(
                (
                    0x00,  # reserved
                    self.user_id,
                )
            )
            + self.message_authentication_code
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            f"<SessionAuthenticate "
            f'user_id="{self.user_id}" '
            f'message_authentication_code="{self.message_authentication_code.hex()}" />'
        )



"""Example for setting different colors on a RGBW remote value."""

import asyncio

from xknx import XKNX
from xknx.dpt import RGBWColor
from xknx.remote_value import RemoteValueColorRGBW


async def main() -> None:
    """Connect to KNX/IP bus and set different colors."""
    xknx = XKNX()
    await xknx.start()

    rgbw = RemoteValueColorRGBW(
        xknx,
        group_address="1/1/40",
        group_address_state="1/1/41",
        device_name="RGBWLight",
    )

    rgbw.set(RGBWColor(255, 255, 255, 0))  # cold-white
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(0, 0, 0, 255))  # warm-white
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(0, 0, 0, 0))  # off
    await asyncio.sleep(1)

    rgbw.set(RGBWColor(255, 0, 0, 0))  # red
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(0, 255, 0, 0))  # green
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(0, 0, 255, 0))  # blue
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(0, 0, 0, 0))  # off
    await asyncio.sleep(1)

    rgbw.set(RGBWColor(255, 255, 0, 0))
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(0, 255, 255, 0))
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(255, 0, 255, 0))
    await asyncio.sleep(1)
    rgbw.set(RGBWColor(0, 0, 0, 0))  # off
    await asyncio.sleep(1)

    await xknx.stop()


asyncio.run(main())



"""Example for daemon mode within XKNX."""

import asyncio

from xknx import XKNX
from xknx.devices import Device, Switch


def device_updated_cb(device: Device) -> None:
    """Do something with the updated device."""
    print(f"Callback received from {device.name}")


async def main() -> None:
    """Connect to KNX/IP device and listen if a switch was updated via KNX bus."""
    xknx = XKNX(device_updated_cb=device_updated_cb, daemon_mode=True)
    xknx.devices.async_add(Switch(xknx, name="TestOutlet", group_address="1/1/11"))
    await xknx.start()
    # Wait until Ctrl-C was pressed
    await xknx.stop()


asyncio.run(main())



"""Implementation of 3.17 Datapoint Types String."""

from __future__ import annotations

from xknx.exceptions import ConversionError

from .dpt import DPTBase
from .payload import DPTArray, DPTBinary


class DPTString(DPTBase):
    """
    Abstraction for KNX 14 Octet ASCII string.

    DPT 16.000
    """

    payload_type = DPTArray
    payload_length = 14
    dpt_main_number = 16
    dpt_sub_number = 0
    value_type = "string"
    unit = None

    _encoding = "ascii"

    @classmethod
    def from_knx(cls, payload: DPTArray | DPTBinary) -> str:
        """Parse/deserialize from KNX/IP raw data."""
        raw = cls.validate_payload(payload)
        return bytes(byte for byte in raw if byte != 0x00).decode(
            cls._encoding, errors="replace"
        )

    @classmethod
    def to_knx(cls, value: str) -> DPTArray:
        """Serialize to KNX/IP raw data."""
        try:
            knx_value = str(value)
            if not cls._test_boundaries(knx_value):
                raise ValueError("Value out of range")
        except ValueError as err:
            raise ConversionError(
                f"Could not serialize {cls.dpt_name()}", value=value
            ) from err
        # replace invalid characters with question marks
        raw_bytes = knx_value.encode(cls._encoding, errors="replace")
        padding = bytes(cls.payload_length - len(raw_bytes))
        return DPTArray(raw_bytes + padding)

    @classmethod
    def _test_boundaries(cls, value: str) -> bool:
        """Test if value is within defined range for this object."""
        return len(value) <= cls.payload_length


class DPTLatin1(DPTString):
    """
    Abstraction for KNX 14 Octet Latin-1 (ISO 8859-1) string.

    DPT 16.001
    """

    dpt_main_number = 16
    dpt_sub_number = 1
    value_type = "latin_1"
    _encoding = "latin_1"



"""Module for managing a text via KNX."""

from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING

from xknx.dpt import DPTString
from xknx.remote_value import GroupAddressesType, RemoteValueString
from xknx.typing import DPTParsable

from .device import Device, DeviceCallbackType

if TYPE_CHECKING:
    from xknx.telegram import Telegram
    from xknx.xknx import XKNX


class Notification(Device):
    """Class for managing a notification."""

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        group_address: GroupAddressesType = None,
        group_address_state: GroupAddressesType = None,
        respond_to_read: bool = False,
        sync_state: bool | int | float | str = True,
        value_type: DPTParsable | type[DPTString] | None = None,
        device_updated_cb: DeviceCallbackType[Notification] | None = None,
    ) -> None:
        """Initialize notification class."""
        super().__init__(xknx, name, device_updated_cb)

        self.respond_to_read = respond_to_read
        self.remote_value = RemoteValueString(
            xknx,
            group_address=group_address,
            group_address_state=group_address_state,
            sync_state=sync_state,
            value_type=value_type,
            device_name=name,
            feature_name="Text",
            after_update_cb=self.after_update,
        )

    def _iter_remote_values(self) -> Iterator[RemoteValueString]:
        """Iterate the devices RemoteValue classes."""
        yield self.remote_value

    @property
    def message(self) -> str | None:
        """Return the current message."""
        return self.remote_value.value

    async def set(self, message: str) -> None:
        """Set message."""
        cropped_message = message[:14]
        self.remote_value.set(cropped_message)

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming and outgoing GROUP WRITE telegram."""
        self.remote_value.process(telegram)

    def process_group_read(self, telegram: Telegram) -> None:
        """Process incoming GroupValueResponse telegrams."""
        if (
            self.respond_to_read
            and telegram.destination_address == self.remote_value.group_address
        ):
            self.remote_value.respond()

    def __str__(self) -> str:
        """Return object as readable string."""
        return f'<Notification name="{self.name}" message={self.remote_value.group_addr_str()} />'



"""Unit test for KNX/IP base class."""

import pytest

from xknx.exceptions import CouldNotParseKNXIP, IncompleteKNXIPFrame
from xknx.knxip import KNXIPFrame, KNXIPHeader
from xknx.knxip.knxip_enum import KNXIPServiceType


class TestKNXIPFrame:
    """Test class for KNX/IP base class."""

    def test_wrong_init(self) -> None:
        """Testing init method with wrong service_type_ident."""
        header = KNXIPHeader()
        header.service_type_ident = KNXIPServiceType.REMOTE_DIAG_RESPONSE
        with pytest.raises(CouldNotParseKNXIP):
            # this is not yet implemented in xknx
            KNXIPFrame.from_knx(header.to_knx())

    def test_double_frame(self) -> None:
        """Test parsing KNX/IP frame from streaming data containing two frames."""
        raw = bytes.fromhex(
            "06 10 04 20 00 15 04 02 51 00 29 00 bc e0 10 fa"
            "09 2d 01 00 80"
            "06 10 04 20 00 15 04 02 52 00 29 00 bc e0 10 1f"
            "08 2d 01 00 80"
        )  # both frames have length 21
        frame_1, rest_1 = KNXIPFrame.from_knx(raw)
        frame_2, _rest_2 = KNXIPFrame.from_knx(rest_1)
        assert frame_1.header.total_length == 21
        assert frame_2.header.total_length == 21

    def test_parsing_too_short_knxip(self) -> None:
        """Test parsing and streaming connection state request KNX/IP packet."""
        raw = bytes.fromhex("06 10 02 07 00 10 15 00 08 01 C0 A8 C8 0C C3")
        with pytest.raises(IncompleteKNXIPFrame):
            KNXIPFrame.from_knx(raw)



"""
Module for managing a DTP 5010 remote value.

DPT 5.010.
"""

from __future__ import annotations

from xknx.dpt import DPTValue1Ucount

from .remote_value import RemoteValue


class RemoteValueDptValue1Ucount(RemoteValue[int]):
    """Abstraction for remote value of KNX DPT 5.010."""

    dpt_class = DPTValue1Ucount



"""
TCPTransport is an abstraction for handling the complete TCP io.

The module is build upon asyncio stream socket functions.
"""

from __future__ import annotations

import asyncio
from collections.abc import Callable
import logging

from xknx.exceptions import CommunicationError, CouldNotParseKNXIP, IncompleteKNXIPFrame
from xknx.knxip import HPAI, HostProtocol, KNXIPFrame

from .ip_transport import KNXIPTransport

raw_socket_logger = logging.getLogger("xknx.raw_socket")
logger = logging.getLogger("xknx.log")
knx_logger = logging.getLogger("xknx.knx")


class TCPTransport(KNXIPTransport):
    """Class for handling (sending and receiving) TCP packets."""

    class TCPTransportFactory(asyncio.Protocol):
        """Abstraction for managing the asyncio-tcp protocol."""

        __slots__ = ("connection_lost_callback", "data_received_callback", "transport")

        def __init__(
            self,
            data_received_callback: Callable[[bytes], None],
            connection_lost_callback: Callable[[], None],
        ) -> None:
            """Initialize UDPTransportFactory class."""
            self.transport: asyncio.BaseTransport | None = None
            self.data_received_callback = data_received_callback
            self.connection_lost_callback = connection_lost_callback

            # Workaround for issue of TCP Transport in ProactorEventLoop in py3.10 and py3.11
            # on Windows returning bytearray instead of bytes which lead to errors in
            # cryptography (eg. X25519PublicKey.from_public_bytes() in IP Secure handshake)
            # https://github.com/python/cpython/issues/99941
            try:
                if isinstance(asyncio.get_event_loop(), asyncio.ProactorEventLoop):  # type: ignore[attr-defined, unused-ignore] # unused-ignore for Windows
                    self.data_received_callback = lambda data: data_received_callback(
                        bytes(data)
                    )
            except AttributeError:
                # asyncio.ProactorEventLoop is only available on Windows
                pass

        def connection_made(self, transport: asyncio.BaseTransport) -> None:
            """Assign transport. Callback after udp connection was made."""
            self.transport = transport

        def data_received(self, data: bytes) -> None:
            """Call assigned callback. Callback for datagram received."""
            raw_socket_logger.debug("Received via tcp: %s", data.hex())
            self.data_received_callback(data)

        def connection_lost(self, exc: Exception | None) -> None:
            """Log error. Callback for connection lost."""
            logger.debug("Closing TCP transport. %s", exc)
            self.connection_lost_callback()

    __slots__ = ("_buffer", "_connection_lost_cb", "remote_hpai")

    def __init__(
        self,
        remote_addr: tuple[str, int],
        connection_lost_cb: Callable[[], None] | None = None,
    ) -> None:
        """Initialize TCPTransport class."""
        self.remote_addr = remote_addr
        self.remote_hpai = HPAI(*remote_addr, protocol=HostProtocol.IPV4_TCP)

        self.callbacks = []
        self._connection_lost_cb = connection_lost_cb
        self.transport: asyncio.Transport | None = None
        self._buffer = b""

    def data_received_callback(self, raw: bytes) -> None:
        """Parse and process KNXIP frame. Callback for having received data over TCP."""
        if self._buffer:
            raw = self._buffer + raw
            self._buffer = b""
        if not raw:
            return
        try:
            knxipframe, next_frame_part = KNXIPFrame.from_knx(raw)
        except IncompleteKNXIPFrame:
            self._buffer = raw
            raw_socket_logger.debug(
                "Incomplete KNX/IP frame. Waiting for rest: %s", raw.hex()
            )
            return
        except CouldNotParseKNXIP as couldnotparseknxip:
            knx_logger.debug(
                "Unsupported KNXIPFrame from %s: %s in %s",
                self.remote_hpai,
                couldnotparseknxip.description,
                raw.hex(),
            )
        else:
            knx_logger.debug(
                "Received from %s: %s",
                self.remote_hpai,
                knxipframe,
            )
            self.handle_knxipframe(knxipframe, self.remote_hpai)
        # parse data after current KNX/IP frame
        if next_frame_part:
            self.data_received_callback(next_frame_part)

    async def connect(self) -> None:
        """Connect TCP socket."""
        tcp_transport_factory = TCPTransport.TCPTransportFactory(
            data_received_callback=self.data_received_callback,
            connection_lost_callback=self._connection_lost,
        )
        loop = asyncio.get_running_loop()
        (self.transport, _) = await loop.create_connection(
            lambda: tcp_transport_factory,
            host=self.remote_hpai.ip_addr,
            port=self.remote_hpai.port,
        )

    def _connection_lost(self) -> None:
        """Call assigned callback. Callback for connection lost."""
        # avoid calling the callback when the transport was stopped intentionally
        if self.transport is not None:
            self.stop()
            if self._connection_lost_cb:
                self._connection_lost_cb()

    def send(self, knxipframe: KNXIPFrame, addr: tuple[str, int] | None = None) -> None:
        """Send KNXIPFrame to socket. `addr` is ignored on TCP."""
        knx_logger.debug(
            "Sending to %s: %s",
            self.remote_hpai,
            knxipframe,
        )
        if self.transport is None:
            raise CommunicationError("Transport not connected")

        self.transport.write(knxipframe.to_knx())



"""Unit test for KNX/IP SessionResponse objects."""

from xknx.knxip import KNXIPFrame, SessionResponse


class TestKNXIPSessionResponse:
    """Test class for KNX/IP SessionResponse objects."""

    def test_session_response(self) -> None:
        """Test parsing and streaming session response KNX/IP packet."""
        public_key = bytes.fromhex(
            "bd f0 99 90 99 23 14 3e"  # Diffie-Hellman Server Public Value Y
            "f0 a5 de 0b 3b e3 68 7b"
            "c5 bd 3c f5 f9 e6 f9 01"
            "69 9c d8 70 ec 1f f8 24"
        )
        message_authentication_code = bytes.fromhex(
            "a9 22 50 5a aa 43 61 63"  # Message Authentication Code
            "57 0b d5 49 4c 2d f2 a3"
        )
        raw = (
            bytes.fromhex(
                "06 10 09 52 00 38"  # KNXnet/IP header
                "00 01"  # Secure Session Identifier
            )
            + public_key
            + message_authentication_code
        )
        knxipframe, _ = KNXIPFrame.from_knx(raw)

        assert isinstance(knxipframe.body, SessionResponse)
        assert knxipframe.body.secure_session_id == 1
        assert knxipframe.body.ecdh_server_public_key == public_key
        assert (
            knxipframe.body.message_authentication_code == message_authentication_code
        )

        assert knxipframe.to_knx() == raw

        session_response = SessionResponse(
            secure_session_id=1,
            ecdh_server_public_key=public_key,
            message_authentication_code=message_authentication_code,
        )
        knxipframe2 = KNXIPFrame.init_from_body(session_response)

        assert knxipframe2.to_knx() == raw



"""
Module for Serialization and Deserialization of KNX Secure Wrapper.

When KNXnet/IP frames are to be sent over a secured connection, each frame including
the KNXnet/IP header shall be completely encapsulated as encrypted payload inside a
SECURE_WRAPPER frame that adds some extra information needed to decrypt the frame and
for ensuring data integrity and freshness.
"""

from __future__ import annotations

from typing import Final

from xknx.exceptions import CouldNotParseKNXIP

from .body import KNXIPBody
from .knxip_enum import KNXIPServiceType

# 2 octets secure session identifier
# 6 octets sequence information
# 6 octets KNX serial number
# 2 octets message tag
SECURITY_INFORMATION_LENGTH: Final = 16
# 6 octets for KNX/IP header
# 2 for smallest payload (eg. SessionStatus)
MINIMUM_PAYLOAD_LENGTH: Final = 2
MESSAGE_AUTHENTICATION_CODE_LENGTH: Final = 16

SECURE_WRAPPER_MINIMUM_LENGTH: Final = (
    SECURITY_INFORMATION_LENGTH
    + MINIMUM_PAYLOAD_LENGTH
    + MESSAGE_AUTHENTICATION_CODE_LENGTH
)


class SecureWrapper(KNXIPBody):
    """Representation of a KNX Secure Wrapper."""

    SERVICE_TYPE = KNXIPServiceType.SECURE_WRAPPER

    def __init__(
        self,
        secure_session_id: int = 0,
        sequence_information: bytes = bytes(6),
        serial_number: bytes = bytes(6),
        message_tag: bytes = bytes(2),
        encrypted_data: bytes = bytes(0),
        message_authentication_code: bytes = bytes(16),
    ) -> None:
        """Initialize SecureWrapper object."""
        self.secure_session_id = secure_session_id
        self.sequence_information = sequence_information
        self.serial_number = serial_number
        self.message_tag = message_tag
        self.encrypted_data = encrypted_data
        self.message_authentication_code = message_authentication_code

    def calculated_length(self) -> int:
        """Get length of KNX/IP body."""
        return (
            SECURITY_INFORMATION_LENGTH
            + len(self.encrypted_data)
            + MESSAGE_AUTHENTICATION_CODE_LENGTH
        )

    def from_knx(self, raw: bytes) -> int:
        """Parse/deserialize from KNX/IP raw data."""
        if len(raw) < SECURE_WRAPPER_MINIMUM_LENGTH:
            raise CouldNotParseKNXIP("SecureWrapper has invalid length")
        self.secure_session_id = int.from_bytes(raw[:2], "big")
        self.sequence_information = raw[2:8]
        self.serial_number = raw[8:14]
        self.message_tag = raw[14:16]
        self.encrypted_data = raw[16:-MESSAGE_AUTHENTICATION_CODE_LENGTH]
        self.message_authentication_code = raw[-MESSAGE_AUTHENTICATION_CODE_LENGTH:]
        return len(raw)

    def to_knx(self) -> bytes:
        """Serialize to KNX/IP raw data."""
        return (
            self.secure_session_id.to_bytes(2, "big")
            + self.sequence_information
            + self.serial_number
            + self.message_tag
            + self.encrypted_data
            + self.message_authentication_code
        )

    def __repr__(self) -> str:
        """Return object as readable string."""
        return (
            f"<SecureWrapper "
            f'secure_session_id="{self.secure_session_id}" '
            f'sequence_information="{self.sequence_information.hex()}" '
            f'serial_number="{self.serial_number.hex()}" '
            f'message_tag="{self.message_tag.hex()}" '
            f'encrypted_data="{self.encrypted_data.hex()}" '
            f'message_authentication_code="{self.message_authentication_code.hex()}" />'
        )



"""Test DPT lookup."""

# flake8: noqa
import pytest

from xknx.dpt import *  # pylint: disable=wildcard-import,unused-wildcard-import

# """
# Generate a list of all DPTs for the following test.
# Run this function manually to update the list.
# """

# from xknx.dpt import DPTBase

# dpt_dict = {}
# for dpt in DPTBase.dpt_class_tree():
#     if dpt.value_type is not None:
#         dpt_dict[dpt.value_type] = dpt

# for _, dpt in sorted(dpt_dict.items()):
#     unit = None
#     if dpt.unit is not None:
#         unit = f'"{dpt.unit}"'
#     print(
#         f'("{dpt.value_type}", {dpt.__name__}, {dpt.dpt_main_number}, {dpt.dpt_sub_number}, {unit}),'
#     )


@pytest.mark.parametrize(
    ("value_type", "dpt_class", "main", "sub", "unit"),
    [
        ("1byte_signed", DPTSignedRelativeValue, 6, None, None),
        ("1byte_unsigned", DPTValue1ByteUnsigned, 5, None, None),
        ("2byte_float", DPT2ByteFloat, 9, None, None),
        ("2byte_signed", DPT2ByteSigned, 8, None, None),
        ("2byte_unsigned", DPT2ByteUnsigned, 7, None, None),
        ("4byte_float", DPT4ByteFloat, 14, None, None),
        ("4byte_signed", DPT4ByteSigned, 13, None, None),
        ("4byte_unsigned", DPT4ByteUnsigned, 12, None, None),
        ("8byte_signed", DPT8ByteSigned, 29, None, None),
        ("absolute_humidity", DPTAbsoluteHumidity, 9, 29, "g/m³"),
        ("absolute_temperature", DPTAbsoluteTemperature, 14, 69, "K"),
        ("acceleration", DPTAcceleration, 14, 0, "m/s²"),
        ("acceleration_angular", DPTAccelerationAngular, 14, 1, "rad/s²"),
        ("ack", DPTAck, 1, 16, None),
        ("activation_energy", DPTActivationEnergy, 14, 2, "J/mol"),
        ("active_energy", DPTActiveEnergy, 13, 10, "Wh"),
        ("active_energy_8byte", DPTActiveEnergy8Byte, 29, 10, "Wh"),
        ("active_energy_kwh", DPTActiveEnergykWh, 13, 13, "kWh"),
        ("active_energy_mwh", DPTActiveEnergyMWh, 13, 16, "MWh"),
        ("activity", DPTActivity, 14, 3, "s⁻¹"),
        ("air_flow", DPTAirFlow, 9, 9, "m³/h"),
        ("alarm", DPTAlarm, 1, 5, None),
        ("amplitude", DPTAmplitude, 14, 5, None),
        ("angle", DPTAngle, 5, 3, "°"),
        ("angle_deg", DPTAngleDeg, 14, 7, "°"),
        ("angle_rad", DPTAngleRad, 14, 6, "rad"),
        ("angular_frequency", DPTAngularFrequency, 14, 34, "rad/s"),
        ("angular_momentum", DPTAngularMomentum, 14, 8, "J s"),
        ("angular_velocity", DPTAngularVelocity, 14, 9, "rad/s"),
        ("apparant_energy", DPTApparantEnergy, 13, 11, "VAh"),
        ("apparant_energy_8byte", DPTApparantEnergy8Byte, 29, 11, "VAh"),
        ("apparant_energy_kvah", DPTApparantEnergykVAh, 13, 14, "kVAh"),
        ("apparent_power", DPTApparentPower, 14, 80, "VA"),
        ("area", DPTArea, 14, 10, "m²"),
        ("binary_value", DPTBinaryValue, 1, 6, None),
        ("bool", DPTBool, 1, 2, None),
        ("brightness", DPTBrightness, 7, 13, "lx"),
        ("capacitance", DPTCapacitance, 14, 11, "F"),
        ("charge_density_surface", DPTChargeDensitySurface, 14, 12, "C/m²"),
        ("charge_density_volume", DPTChargeDensityVolume, 14, 13, "C/m³"),
        ("color_rgb", DPTColorRGB, 232, 600, None),
        ("color_rgbw", DPTColorRGBW, 251, 600, None),
        ("color_temperature", DPTColorTemperature, 7, 600, "K"),
        ("color_xyy", DPTColorXYY, 242, 600, None),
        ("common_temperature", DPTCommonTemperature, 14, 68, "°C"),
        ("compressibility", DPTCompressibility, 14, 14, "m²/N"),
        ("concentration_ugm3", DPTConcentrationUGM3, 9, 30, "μg/m³"),
        ("conductance", DPTConductance, 14, 15, "S"),
        ("consumer_producer", DPTConsumerProducer, 1, 1200, None),
        ("control_blinds", DPTControlBlinds, 3, 8, None),
        ("control_dimming", DPTControlDimming, 3, 7, None),
        ("counter_pulses", DPTValue1Count, 6, 10, "counter pulses"),
        ("curr", DPTCurrent, 9, 21, "mA"),
        ("current", DPTUElCurrentmA, 7, 12, "mA"),
        ("date", DPTDate, 11, 1, None),
        ("datetime", DPTDateTime, 19, 1, None),
        ("day_night", DPTDayNight, 1, 24, None),
        ("decimal_factor", DPTDecimalFactor, 5, 5, None),
        ("delta_time_100ms", DPTDeltaTime100Msec, 8, 4, "ms"),
        ("delta_time_10ms", DPTDeltaTime10Msec, 8, 3, "ms"),
        ("delta_time_hrs", DPTDeltaTimeHrs, 8, 7, "h"),
        ("delta_time_min", DPTDeltaTimeMin, 8, 6, "min"),
        ("delta_time_ms", DPTDeltaTimeMsec, 8, 2, "ms"),
        ("delta_time_sec", DPTDeltaTimeSec, 8, 5, "s"),
        ("delta_volume_liquid_litre", DPTDeltaVolumeLiquidLitre, 13, 1200, "L"),
        ("delta_volume_m3", DPTDeltaVolumeM3, 13, 1201, "m³"),
        ("density", DPTDensity, 14, 17, "kg/m³"),
        ("dim_send_style", DPTDimSendStyle, 1, 13, None),
        ("electric_charge", DPTElectricCharge, 14, 18, "C"),
        ("electric_current", DPTElectricCurrent, 14, 19, "A"),
        ("electric_current_density", DPTElectricCurrentDensity, 14, 20, "A/m²"),
        ("electric_dipole_moment", DPTElectricDipoleMoment, 14, 21, "C m"),
        ("electric_displacement", DPTElectricDisplacement, 14, 22, "C/m²"),
        ("electric_field_strength", DPTElectricFieldStrength, 14, 23, "V/m"),
        ("electric_flux", DPTElectricFlux, 14, 24, "c"),
        ("electric_flux_density", DPTElectricFluxDensity, 14, 25, "C/m²"),
        ("electric_polarization", DPTElectricPolarization, 14, 26, "C/m²"),
        ("electric_potential", DPTElectricPotential, 14, 27, "V"),
        ("electric_potential_difference", DPTElectricPotentialDifference, 14, 28, "V"),
        ("electrical_conductivity", DPTElectricalConductivity, 14, 16, "S/m"),
        ("electromagnetic_moment", DPTElectromagneticMoment, 14, 29, "A m²"),
        ("electromotive_force", DPTElectromotiveForce, 14, 30, "V"),
        ("enable", DPTEnable, 1, 3, None),
        ("energy", DPTEnergy, 14, 31, "J"),
        ("energy_direction", DPTEnergyDirection, 1, 1201, None),
        ("enthalpy", DPTEnthalpy, 9, 60000, "H"),
        ("flow_rate_m3h", DPTFlowRateM3H, 13, 2, "m³/h"),
        ("force", DPTForce, 14, 32, "N"),
        ("frequency", DPTFrequency, 14, 33, "Hz"),
        ("heat_cool", DPTHeatCool, 1, 100, None),
        ("heat_quantity", DPTHeatQuantity, 14, 37, "J"),
        ("heatcapacity", DPTHeatCapacity, 14, 35, "J/K"),
        ("heatflowrate", DPTHeatFlowRate, 14, 36, "W"),
        ("humidity", DPTHumidity, 9, 7, "%"),
        ("hvac_controller_mode", DPTHVACContrMode, 20, 105, None),
        ("hvac_mode", DPTHVACMode, 20, 102, None),
        ("hvac_status", DPTHVACStatus, 20, 60102, None),
        ("illuminance", DPTLux, 9, 4, "lx"),
        ("impedance", DPTImpedance, 14, 38, "Ω"),
        ("input_source", DPTInputSource, 1, 14, None),
        ("invert", DPTInvert, 1, 12, None),
        ("kelvin_per_percent", DPTKelvinPerPercent, 9, 23, "K/%"),
        ("latin_1", DPTLatin1, 16, 1, None),
        ("length", DPTLength, 14, 39, "m"),
        ("length_m", DPTLengthM, 8, 12, "m"),
        ("length_mm", DPTLengthMm, 7, 11, "mm"),
        ("light_quantity", DPTLightQuantity, 14, 40, "lm s"),
        ("logical_function", DPTLogicalFunction, 1, 21, None),
        ("long_delta_timesec", DPTLongDeltaTimeSec, 13, 100, "s"),
        ("long_time_period_hrs", DPTLongTimePeriodHrs, 12, 102, "h"),
        ("long_time_period_min", DPTLongTimePeriodMin, 12, 101, "min"),
        ("long_time_period_sec", DPTLongTimePeriodSec, 12, 100, "s"),
        ("luminance", DPTLuminance, 14, 41, "cd/m²"),
        ("luminous_flux", DPTLuminousFlux, 14, 42, "lm"),
        ("luminous_intensity", DPTLuminousIntensity, 14, 43, "cd"),
        ("magnetic_field_strength", DPTMagneticFieldStrength, 14, 44, "A/m"),
        ("magnetic_flux", DPTMagneticFlux, 14, 45, "Wb"),
        ("magnetic_flux_density", DPTMagneticFluxDensity, 14, 46, "T"),
        ("magnetic_moment", DPTMagneticMoment, 14, 47, "A m²"),
        ("magnetic_polarization", DPTMagneticPolarization, 14, 48, "T"),
        ("magnetization", DPTMagnetization, 14, 49, "A/m"),
        ("magnetomotive_force", DPTMagnetomotiveForce, 14, 50, "A"),
        ("mass", DPTMass, 14, 51, "kg"),
        ("mass_flux", DPTMassFlux, 14, 52, "kg/s"),
        ("mol", DPTMol, 14, 4, "mol"),
        ("momentum", DPTMomentum, 14, 53, "N/s"),
        ("occupancy", DPTOccupancy, 1, 18, None),
        ("open_close", DPTOpenClose, 1, 9, None),
        ("percent", DPTScaling, 5, 1, "%"),
        ("percentU8", DPTPercentU8, 5, 4, "%"),
        ("percentV16", DPTPercentV16, 8, 10, "%"),
        ("percentV8", DPTPercentV8, 6, 1, "%"),
        ("phaseangledeg", DPTPhaseAngleDeg, 14, 55, "°"),
        ("phaseanglerad", DPTPhaseAngleRad, 14, 54, "rad"),
        ("power", DPTPower, 14, 56, "W"),
        ("power_2byte", DPTPower2Byte, 9, 24, "kW"),
        ("power_density", DPTPowerDensity, 9, 22, "W/m²"),
        ("powerfactor", DPTPowerFactor, 14, 57, None),
        ("ppm", DPTPartsPerMillion, 9, 8, "ppm"),
        ("pressure", DPTPressure, 14, 58, "Pa"),
        ("pressure_2byte", DPTPressure2Byte, 9, 6, "Pa"),
        ("prop_data_type", DPTPropDataType, 7, 10, None),
        ("pulse", DPTValue1Ucount, 5, 10, "counter pulses"),
        ("pulse_2byte", DPT2Ucount, 7, 1, "pulses"),
        ("pulse_2byte_signed", DPTValue2Count, 8, 1, "pulses"),
        ("pulse_4_ucount", DPTValue4Ucount, 12, 1, "counter pulses"),
        ("pulse_4byte", DPTValue4Count, 13, 1, "counter pulses"),
        ("rain_amount", DPTRainAmount, 9, 26, "L/m²"),
        ("ramp", DPTRamp, 1, 4, None),
        ("reactance", DPTReactance, 14, 59, "Ω"),
        ("reactive_energy", DPTReactiveEnergy, 13, 12, "VARh"),
        ("reactive_energy_8byte", DPTReactiveEnergy8Byte, 29, 12, "VARh"),
        ("reactive_energy_kvarh", DPTReactiveEnergykVARh, 13, 15, "kVARh"),
        ("reset", DPTReset, 1, 15, None),
        ("resistance", DPTResistance, 14, 60, "Ω"),
        ("resistivity", DPTResistivity, 14, 61, "Ωm"),
        ("rotation_angle", DPTRotationAngle, 8, 11, "°"),
        ("scene_ab", DPTSceneAB, 1, 22, None),
        ("scene_control", DPTSceneControl, 18, 1, None),
        ("scene_number", DPTSceneNumber, 17, 1, None),
        ("self_inductance", DPTSelfInductance, 14, 62, "H"),
        ("shutter_blinds_mode", DPTShutterBlindsMode, 1, 23, None),
        ("solid_angle", DPTSolidAngle, 14, 63, "sr"),
        ("sound_intensity", DPTSoundIntensity, 14, 64, "W/m²"),
        ("speed", DPTSpeed, 14, 65, "m/s"),
        ("start", DPTStart, 1, 10, None),
        ("state", DPTState, 1, 11, None),
        ("step", DPTStep, 1, 7, None),
        ("stress", DPTStress, 14, 66, "Pa"),
        ("string", DPTString, 16, 0, None),
        ("surface_tension", DPTSurfaceTension, 14, 67, "N/m"),
        ("switch", DPTSwitch, 1, 1, None),
        ("tariff", DPTTariff, 5, 6, None),
        ("tariff_active_energy", DPTTariffActiveEnergy, 235, 1, None),
        ("temperature", DPTTemperature, 9, 1, "°C"),
        ("temperature_a", DPTTemperatureA, 9, 3, "K/h"),
        ("temperature_difference", DPTTemperatureDifference, 14, 70, "K"),
        ("temperature_difference_2byte", DPTTemperatureDifference2Byte, 9, 2, "K"),
        ("temperature_f", DPTTemperatureF, 9, 27, "°F"),
        ("thermal_capacity", DPTThermalCapacity, 14, 71, "J/K"),
        ("thermal_conductivity", DPTThermalConductivity, 14, 72, "W/mK"),
        ("thermoelectric_power", DPTThermoelectricPower, 14, 73, "V/K"),
        ("time", DPTTime, 10, 1, None),
        ("time_1", DPTTime1, 9, 10, "s"),
        ("time_2", DPTTime2, 9, 11, "ms"),
        ("time_period_100msec", DPTTimePeriod100Msec, 7, 4, "ms"),
        ("time_period_10msec", DPTTimePeriod10Msec, 7, 3, "ms"),
        ("time_period_hrs", DPTTimePeriodHrs, 7, 7, "h"),
        ("time_period_min", DPTTimePeriodMin, 7, 6, "min"),
        ("time_period_msec", DPTTimePeriodMsec, 7, 2, "ms"),
        ("time_period_sec", DPTTimePeriodSec, 7, 5, "s"),
        ("time_seconds", DPTTimeSeconds, 14, 74, "s"),
        ("torque", DPTTorque, 14, 75, "Nm"),
        ("trigger", DPTTrigger, 1, 17, None),
        ("up_down", DPTUpDown, 1, 8, None),
        ("voltage", DPTVoltage, 9, 20, "mV"),
        ("volume", DPTVolume, 14, 76, "m³"),
        ("volume_flow", DPTVolumeFlow, 9, 25, "L/h"),
        ("volume_flux", DPTVolumeFlux, 14, 77, "m³/s"),
        ("volume_flux_ls", DPTVolumeFluxLs, 14, 1201, "L/s"),
        ("volume_flux_meter", DPTVolumeFluxMeter, 14, 1200, "m³/h"),
        ("volume_liquid_litre", DPTVolumeLiquidLitre, 12, 1200, "L"),
        ("volume_m3", DPTVolumeM3, 12, 1201, "m³"),
        ("weight", DPTWeight, 14, 78, "N"),
        ("wind_speed_kmh", DPTWspKmh, 9, 28, "km/h"),
        ("wind_speed_ms", DPTWsp, 9, 5, "m/s"),
        ("window_door", DPTWindowDoor, 1, 19, None),
        ("work", DPTWork, 14, 79, "J"),
    ],
)
def test_dpt_lookup(
    value_type: str,
    dpt_class: type[DPTBase],
    main: int,
    sub: int | None,
    unit: str | None,
) -> None:
    """Test DPT4ByteUnsigned."""
    dpt = DPTBase.parse_transcoder(value_type)
    assert dpt is dpt_class
    assert DPTBase.transcoder_by_dpt(dpt_main=main, dpt_sub=sub) is dpt_class
    assert dpt.unit == unit



"""Unit test for KNX DPT 1."""

from typing import Any

import pytest

from xknx.dpt import DPTArray, DPTBinary, DPTEnumData
from xknx.dpt.dpt_1 import (
    Ack,
    Alarm,
    BinaryValue,
    Bool,
    ConsumerProducer,
    DayNight,
    DimSendStyle,
    DPT1BitEnum,
    DPTAck,
    DPTAlarm,
    DPTBinaryValue,
    DPTBool,
    DPTConsumerProducer,
    DPTDayNight,
    DPTDimSendStyle,
    DPTEnable,
    DPTEnergyDirection,
    DPTHeatCool,
    DPTInputSource,
    DPTInvert,
    DPTLogicalFunction,
    DPTOccupancy,
    DPTOpenClose,
    DPTRamp,
    DPTReset,
    DPTSceneAB,
    DPTShutterBlindsMode,
    DPTStart,
    DPTState,
    DPTStep,
    DPTSwitch,
    DPTTrigger,
    DPTUpDown,
    DPTWindowDoor,
    Enable,
    EnergyDirection,
    HeatCool,
    InputSource,
    Invert,
    LogicalFunction,
    Occupancy,
    OpenClose,
    Ramp,
    Reset,
    SceneAB,
    ShutterBlindsMode,
    Start,
    State,
    Step,
    Switch,
    Trigger,
    UpDown,
    WindowDoor,
)
from xknx.exceptions import ConversionError, CouldNotParseTelegram


@pytest.mark.parametrize(
    ("dpt", "value_false", "value_true"),
    [
        (DPTSwitch, Switch.OFF, Switch.ON),
        (DPTBool, Bool.FALSE, Bool.TRUE),
        (DPTEnable, Enable.DISABLE, Enable.ENABLE),
        (DPTRamp, Ramp.NO_RAMP, Ramp.RAMP),
        (DPTAlarm, Alarm.NO_ALARM, Alarm.ALARM),
        (DPTBinaryValue, BinaryValue.LOW, BinaryValue.HIGH),
        (DPTStep, Step.DECREASE, Step.INCREASE),
        (DPTUpDown, UpDown.UP, UpDown.DOWN),
        (DPTOpenClose, OpenClose.OPEN, OpenClose.CLOSE),
        (DPTStart, Start.STOP, Start.START),
        (DPTState, State.INACTIVE, State.ACTIVE),
        (DPTInvert, Invert.NOT_INVERTED, Invert.INVERTED),
        (DPTDimSendStyle, DimSendStyle.START_STOP, DimSendStyle.CYCLICALLY),
        (DPTInputSource, InputSource.FIXED, InputSource.CALCULATED),
        (DPTReset, Reset.NO_ACTION, Reset.RESET),
        (DPTAck, Ack.NO_ACTION, Ack.ACKNOWLEDGE),
        (DPTTrigger, Trigger.TRIGGER_0, Trigger.TRIGGER),
        (DPTOccupancy, Occupancy.NOT_OCCUPIED, Occupancy.OCCUPIED),
        (DPTWindowDoor, WindowDoor.CLOSED, WindowDoor.OPEN),
        (DPTLogicalFunction, LogicalFunction.OR, LogicalFunction.AND),
        (DPTSceneAB, SceneAB.SCENE_A, SceneAB.SCENE_B),
        (
            DPTShutterBlindsMode,
            ShutterBlindsMode.UP_DOWN_MODE,
            ShutterBlindsMode.STEP_STOP_MODE,
        ),
        (DPTDayNight, DayNight.DAY, DayNight.NIGHT),
        (DPTHeatCool, HeatCool.COOL, HeatCool.HEAT),
        (DPTConsumerProducer, ConsumerProducer.CONSUMER, ConsumerProducer.PRODUCER),
        (DPTEnergyDirection, EnergyDirection.POSITIVE, EnergyDirection.NEGATIVE),
    ],
)
class TestDPT1:
    """Test class for KNX DPT 1 values."""

    def test_to_knx(
        self,
        dpt: type[DPT1BitEnum[Any]],
        value_true: DPTEnumData,
        value_false: DPTEnumData,
    ) -> None:
        """Test parsing to KNX."""
        assert dpt.to_knx(value_true) == DPTBinary(1)
        assert dpt.to_knx(value_false) == DPTBinary(0)

    def test_to_knx_by_string(
        self,
        dpt: type[DPT1BitEnum[Any]],
        value_true: DPTEnumData,
        value_false: DPTEnumData,
    ) -> None:
        """Test parsing string values to KNX."""
        assert dpt.to_knx(value_true.name.lower()) == DPTBinary(1)
        assert dpt.to_knx(value_false.name.lower()) == DPTBinary(0)

    def test_to_knx_by_value(
        self,
        dpt: type[DPT1BitEnum[Any]],
        value_true: DPTEnumData,
        value_false: DPTEnumData,
    ) -> None:
        """Test parsing string values to KNX."""
        assert dpt.to_knx(True) == DPTBinary(1)
        assert dpt.to_knx(1) == DPTBinary(1)
        assert dpt.to_knx(False) == DPTBinary(0)
        assert dpt.to_knx(0) == DPTBinary(0)

    def test_to_knx_wrong_value(
        self,
        dpt: type[DPT1BitEnum[Any]],
        value_true: DPTEnumData,
        value_false: DPTEnumData,
    ) -> None:
        """Test serializing to KNX with wrong value."""
        with pytest.raises(ConversionError):
            dpt.to_knx(2)

    def test_from_knx(
        self,
        dpt: type[DPT1BitEnum[Any]],
        value_true: DPTEnumData,
        value_false: DPTEnumData,
    ) -> None:
        """Test parsing from KNX."""
        assert dpt.from_knx(DPTBinary(0)) == value_false
        assert dpt.from_knx(DPTBinary(1)) == value_true

    def test_from_knx_wrong_value(
        self,
        dpt: type[DPT1BitEnum[Any]],
        value_true: DPTEnumData,
        value_false: DPTEnumData,
    ) -> None:
        """Test parsing with wrong value)."""
        with pytest.raises(CouldNotParseTelegram):
            dpt.from_knx(DPTArray((1,)))
        with pytest.raises(CouldNotParseTelegram):
            dpt.from_knx(DPTBinary(2))



"""Unit test for IP Secure primitives."""

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey

from xknx.secure.security_primitives import (
    calculate_message_authentication_code_cbc,
    decrypt_ctr,
    derive_device_authentication_password,
    derive_user_password,
    encrypt_data_ctr,
    generate_ecdh_key_pair,
)


class TestIPSecure:
    """Test class for IP Secure primitives."""

    def test_calculate_message_authentication_code_cbc(self) -> None:
        """Test calculate message authentication code CBC."""
        # SessionResponse from example in KNX specification AN159v06
        assert calculate_message_authentication_code_cbc(
            key=derive_device_authentication_password("trustme"),
            additional_data=bytes.fromhex(
                "06 10 09 52 00 38 00 01 b7 52 be 24 64 59 26 0f"
                "6b 0c 48 01 fb d5 a6 75 99 f8 3b 40 57 b3 ef 1e"
                "79 e4 69 ac 17 23 4e 15"
            ),
        ) == bytes.fromhex("da 3d c6 af 79 89 6a a6 ee 75 73 d6 99 50 c2 83")
        # RoutingIndication from example in KNX specification AN159v06
        assert calculate_message_authentication_code_cbc(
            key=bytes.fromhex("00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f"),
            additional_data=bytes.fromhex("06 10 09 50 00 37 00 00"),
            payload=bytes.fromhex("06 10 05 30 00 11 29 00 bc d0 11 59 0a de 01 00 81"),
            block_0=bytes.fromhex("c0 c1 c2 c3 c4 c5 00 fa 12 34 56 78 af fe 00 11"),
        ) == bytes.fromhex("bd 0a 29 4b 95 25 54 b2 35 39 20 4c 22 71 d2 6b")

    def test_encrypt_data_ctr(self) -> None:
        """Test encrypt data with AES-CTR."""
        # RoutingIndication from example in KNX specification AN159v06
        key = bytes.fromhex("00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f")
        counter_0 = bytes.fromhex("c0 c1 c2 c3 c4 c5 00 fa 12 34 56 78 af fe ff 00")
        mac_cbc = bytes.fromhex("bd 0a 29 4b 95 25 54 b2 35 39 20 4c 22 71 d2 6b")
        payload = bytes.fromhex("06 10 05 30 00 11 29 00 bc d0 11 59 0a de 01 00 81")
        encrypted_data, mac = encrypt_data_ctr(
            key=key, counter_0=counter_0, mac_cbc=mac_cbc, payload=payload
        )
        assert encrypted_data == bytes.fromhex(
            "b7 ee 7e 8a 1c 2f 7b ba be c7 75 fd 6e 10 d0 bc 4b"
        )
        assert mac == bytes.fromhex("72 12 a0 3a aa e4 9d a8 56 89 77 4c 1d 2b 4d a4")

    def test_decrypt_ctr(self) -> None:
        """Test decrypt data with AES-CTR."""
        # RoutingIndication from example in KNX specification AN159v06
        key = bytes.fromhex("00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f")
        counter_0 = bytes.fromhex("c0 c1 c2 c3 c4 c5 00 fa 12 34 56 78 af fe ff 00")
        mac = bytes.fromhex("72 12 a0 3a aa e4 9d a8 56 89 77 4c 1d 2b 4d a4")
        payload = bytes.fromhex("b7 ee 7e 8a 1c 2f 7b ba be c7 75 fd 6e 10 d0 bc 4b")
        decrypted_data, mac_tr = decrypt_ctr(
            key=key,
            counter_0=counter_0,
            mac=mac,
            payload=payload,
        )
        assert decrypted_data == bytes.fromhex(
            "06 10 05 30 00 11 29 00 bc d0 11 59 0a de 01 00 81"
        )
        assert mac_tr == bytes.fromhex(
            "bd 0a 29 4b 95 25 54 b2 35 39 20 4c 22 71 d2 6b"
        )

    def test_derive_device_authentication_password(self) -> None:
        """Test derive device authentication password."""
        assert derive_device_authentication_password("trustme") == bytes.fromhex(
            "e1 58 e4 01 20 47 bd 6c c4 1a af bc 5c 04 c1 fc"
        )

    def test_derive_user_password(self) -> None:
        """Test derive user password."""
        assert derive_user_password("secret") == bytes.fromhex(
            "03 fc ed b6 66 60 25 1e c8 1a 1a 71 69 01 69 6a"
        )

    def test_generate_ecdh_key_pair(self) -> None:
        """Test generate ECDH key pair."""
        private_key, public_key_bytes = generate_ecdh_key_pair()
        assert isinstance(private_key, X25519PrivateKey)
        assert isinstance(public_key_bytes, bytes)
        assert len(public_key_bytes) == 32
        private_key_2, public_key_bytes_2 = generate_ecdh_key_pair()
        assert private_key != private_key_2
        assert public_key_bytes != public_key_bytes_2



"""Module for communication tools."""

from .group_communication import (
    group_value_read,
    group_value_response,
    group_value_write,
    read_group_value,
)

__all__ = [
    "group_value_read",
    "group_value_response",
    "group_value_write",
    "read_group_value",
]



"""Unit test for NumericValue objects."""

from unittest.mock import Mock

import pytest

from xknx import XKNX
from xknx.devices import NumericValue
from xknx.dpt import DPTArray
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite


class TestNumericValue:
    """Test class for NumericValue objects."""

    @pytest.mark.parametrize(
        "value_type,raw_payload,expected_state",
        [
            # DPT-14 values are according to ETS group monitor values
            (
                "absolute_temperature",
                DPTArray((0x44, 0xD7, 0xD2, 0x8B)),
                1726.579,
            ),
            (
                "angle",
                DPTArray((0xE4,)),
                322,
            ),
            (
                "brightness",
                DPTArray((0xC3, 0x56)),
                50006,
            ),
            (
                "color_temperature",
                DPTArray((0x6C, 0x95)),
                27797,
            ),
            (
                "counter_pulses",
                DPTArray((0x9D,)),
                -99,
            ),
            (
                "current",
                DPTArray((0xCA, 0xCC)),
                51916,
            ),
            (
                "percent",
                DPTArray((0xE3,)),
                89,
            ),
            (
                "percentU8",
                DPTArray((0x6B,)),
                107,
            ),
            (
                "percentV8",
                DPTArray((0x20,)),
                32,
            ),
            (
                "percentV16",
                DPTArray((0x8A, 0x2F)),
                -301.61,
            ),
            (
                "pulse",
                DPTArray((0xFC,)),
                252,
            ),
            (
                "scene_number",
                DPTArray((0x1,)),
                2,
            ),
        ],
    )
    async def test_sensor_value_types(
        self,
        value_type: str,
        raw_payload: DPTArray,
        expected_state: int,
    ) -> None:
        """Test sensor value types."""
        xknx = XKNX()
        sensor = NumericValue(
            xknx,
            "TestSensor",
            group_address="1/2/3",
            value_type=value_type,
        )
        sensor.process(
            Telegram(
                destination_address=GroupAddress("1/2/3"),
                payload=GroupValueWrite(value=raw_payload),
            )
        )

        assert sensor.resolve_state() == expected_state

    #
    # TEST RESPOND
    #
    async def test_respond_to_read(self) -> None:
        """Test respond_to_read function."""
        xknx = XKNX()
        responding = NumericValue(
            xknx,
            "TestSensor1",
            group_address="1/1/1",
            respond_to_read=True,
            value_type="volume_liquid_litre",
        )
        non_responding = NumericValue(
            xknx,
            "TestSensor2",
            group_address="1/1/1",
            respond_to_read=False,
            value_type="volume_liquid_litre",
        )
        responding_multiple = NumericValue(
            xknx,
            "TestSensor3",
            group_address=["1/1/1", "3/3/3"],
            group_address_state="2/2/2",
            respond_to_read=True,
            value_type="volume_liquid_litre",
        )
        #  set initial payload of NumericValue
        responding.sensor_value.value = 256
        non_responding.sensor_value.value = 256
        responding_multiple.sensor_value.value = 256

        read_telegram = Telegram(
            destination_address=GroupAddress("1/1/1"), payload=GroupValueRead()
        )
        # verify no response when respond is False
        non_responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 0

        # verify response when respond is True
        responding.process(read_telegram)
        assert xknx.telegrams.qsize() == 1
        response = xknx.telegrams.get_nowait()
        assert response == Telegram(
            destination_address=GroupAddress("1/1/1"),
            payload=GroupValueResponse(DPTArray((0x00, 0x00, 0x01, 0x00))),
        )
        # verify no response when GroupValueRead request is not for group_address
        responding_multiple.process(read_telegram)
        assert xknx.telegrams.qsize() == 1
        response = xknx.telegrams.get_nowait()
        assert response == Telegram(
            destination_address=GroupAddress("1/1/1"),
            payload=GroupValueResponse(DPTArray((0x00, 0x00, 0x01, 0x00))),
        )
        responding_multiple.process(
            Telegram(
                destination_address=GroupAddress("2/2/2"), payload=GroupValueRead()
            )
        )
        responding_multiple.process(
            Telegram(
                destination_address=GroupAddress("3/3/3"), payload=GroupValueRead()
            )
        )
        assert xknx.telegrams.qsize() == 0

    #
    # TEST SET
    #
    async def test_set_percent(self) -> None:
        """Test set with percent numeric value."""
        xknx = XKNX()
        num_value = NumericValue(
            xknx, "TestSensor", group_address="1/2/3", value_type="percent"
        )
        await num_value.set(75)
        assert xknx.telegrams.qsize() == 1

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0xBF,))),
        )

    async def test_set_temperature(self) -> None:
        """Test set with temperature numeric value."""
        xknx = XKNX()
        num_value = NumericValue(
            xknx, "TestSensor", group_address="1/2/3", value_type="temperature"
        )
        await num_value.set(21.0)
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x0C, 0x1A))),
        )
        # test HomeAssistant device class
        assert num_value.ha_device_class() == "temperature"

    #
    # SYNC
    #
    async def test_sync(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        sensor = NumericValue(
            xknx, "TestSensor", value_type="temperature", group_address_state="1/2/3"
        )
        await sensor.sync()
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )

    #
    # TEST PROCESS
    #
    async def test_process(self) -> None:
        """Test process / reading telegrams from telegram queue."""
        xknx = XKNX()
        sensor = NumericValue(
            xknx, "TestSensor", value_type="temperature", group_address="1/2/3"
        )

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x06, 0xA0))),
        )
        sensor.process(telegram)
        assert sensor.sensor_value.value == 16.96
        assert sensor.sensor_value.telegram.payload.value == DPTArray((0x06, 0xA0))
        assert sensor.resolve_state() == 16.96

    async def test_process_callback(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is called."""

        xknx = XKNX()
        sensor = NumericValue(
            xknx, "TestSensor", group_address="1/2/3", value_type="temperature"
        )
        after_update_callback = Mock()
        sensor.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x01, 0x02))),
        )
        sensor.process(telegram)
        after_update_callback.assert_called_with(sensor)
        assert sensor.last_telegram == telegram
        # consecutive telegrams with same payload shall only trigger one callback
        after_update_callback.reset_mock()
        sensor.process(telegram)
        after_update_callback.assert_not_called()

    async def test_process_callback_always(self) -> None:
        """Test process / reading telegrams from telegram queue. Test if callback is called."""

        xknx = XKNX()
        sensor = NumericValue(
            xknx,
            "TestSensor",
            group_address="1/2/3",
            value_type="temperature",
            always_callback=True,
        )
        after_update_callback = Mock()
        sensor.register_device_updated_cb(after_update_callback)

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x01, 0x02))),
        )
        sensor.process(telegram)
        after_update_callback.assert_called_with(sensor)
        assert sensor.last_telegram == telegram
        # every telegram shall trigger callback
        after_update_callback.reset_mock()
        sensor.process(telegram)
        after_update_callback.assert_called_with(sensor)

    async def test_process_callback_set(self) -> None:
        """Test setting value. Test if callback is called."""

        xknx = XKNX()
        after_update_callback = Mock()
        num_value = NumericValue(
            xknx, "TestSensor", group_address="1/2/3", value_type="temperature"
        )
        xknx.devices.async_add(num_value)
        num_value.register_device_updated_cb(after_update_callback)

        await num_value.set(21.0)
        xknx.devices.process(xknx.telegrams.get_nowait())
        after_update_callback.assert_called_with(num_value)

    def test_string(self) -> None:
        """Test NumericValue string representation."""

        xknx = XKNX()
        value = NumericValue(
            xknx, "Num", group_address="1/2/3", value_type="temperature"
        )
        value.sensor_value.value = 4.9
        assert (
            str(value)
            == '<NumericValue name="Num" addresses=<1/2/3, None, [], 4.9 /> value=4.9 unit="°C"/>'
        )



"""Unit test for DateTime object."""

import datetime as dt
from unittest.mock import AsyncMock, patch
from zoneinfo import ZoneInfo

from freezegun import freeze_time
import pytest

from xknx import XKNX
from xknx.core import XknxConnectionState
from xknx.devices.datetime import (
    BROADCAST_MINUTES,
    DateDevice,
    DateTimeDevice,
    TimeDevice,
)
from xknx.dpt import DPTArray
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite

from ..conftest import EventLoopClockAdvancer


class TestDateTime:
    """Test class for Time object."""

    @pytest.mark.parametrize(
        ("test_cls", "dt_value", "raw"),
        [
            (TimeDevice, dt.time(9, 13, 14), (0x9, 0xD, 0xE)),
            (DateDevice, dt.date(2017, 1, 7), (0x07, 0x01, 0x11)),
            (
                DateTimeDevice,
                dt.datetime(2017, 1, 7, 9, 13, 14),
                (0x75, 0x01, 0x07, 0x09, 0x0D, 0x0E, 0x24, 0x00),
            ),
        ],
    )
    async def test_process_set_custom_time(
        self,
        test_cls: type[DateDevice | DateTimeDevice | TimeDevice],
        dt_value: dt.time | dt.date | dt.datetime,
        raw: tuple[int],
    ) -> None:
        """Test setting a new time."""
        xknx = XKNX()
        test_device = test_cls(
            xknx,
            "Test",
            group_address="1/2/3",
            localtime=False,
        )
        assert test_device.value is None

        await test_device.set(dt_value)
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray(raw)),
        )
        test_device.process(telegram)
        assert test_device.value == dt_value

    @pytest.mark.parametrize(
        ("cls", "raw_length", "raw"),
        [
            (TimeDevice, 3, (0xC9, 0xD, 0xE)),
            (DateDevice, 3, (0x07, 0x01, 0x11)),
            (DateTimeDevice, 8, (0x75, 0x01, 0x07, 0xC9, 0x0D, 0x0E, 0x20, 0xC0)),
        ],
    )
    async def test_sync_localtime(
        self,
        cls: type[DateDevice | DateTimeDevice | TimeDevice],
        raw_length: int,
        raw: tuple[int],
    ) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        test_device = cls(xknx, "Test", group_address="1/2/3")

        with freeze_time("2017-01-07 09:13:14"):
            await test_device.sync()

        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram.destination_address == GroupAddress("1/2/3")
        assert len(telegram.payload.value.value) == raw_length
        assert telegram.payload.value.value == raw

    async def test_sync_time_custom(self) -> None:
        """Test sync function / sending group reads to KNX bus."""
        xknx = XKNX()
        test_device = TimeDevice(
            xknx,
            "TestDateTime",
            group_address="1/2/3",
            group_address_state="1/2/4",
            localtime=False,
        )
        assert test_device.has_group_address(GroupAddress("1/2/4"))
        await test_device.sync()

        telegram = xknx.telegrams.get_nowait()
        assert telegram.destination_address == GroupAddress("1/2/4")
        assert isinstance(telegram.payload, GroupValueRead)

    @pytest.mark.parametrize(
        ("expected", "localtime"),
        [
            ((0xC9, 0xD, 0xE), True),
            ((0xCA, 0xD, 0xE), ZoneInfo("Europe/Vienna")),
        ],
    )
    async def test_process_read_localtime_time(
        self, expected: tuple[int, int, int], localtime: bool | dt.tzinfo
    ) -> None:
        """Test test process a read telegram from KNX bus."""
        xknx = XKNX()
        test_device = TimeDevice(
            xknx, "TestTime", group_address="1/2/3", localtime=localtime
        )

        telegram_read = Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )
        with freeze_time("2017-01-07 09:13:14"):
            test_device.process(telegram_read)

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(DPTArray(expected)),
        )

    @pytest.mark.parametrize(
        ("expected", "localtime"),
        [
            (
                (0x75, 0x07, 0x0B, 0x49, 0x0D, 0x0E, 0x20, 0xC0),
                True,
            ),
            (
                (0x75, 0x07, 0x0B, 0x4B, 0x0D, 0x0E, 0x21, 0xC0),
                ZoneInfo("Europe/Vienna"),
            ),
        ],
    )
    async def test_process_read_localtime_datetime(
        self, expected: tuple[int, int, int], localtime: bool | dt.tzinfo
    ) -> None:
        """Test test process a read telegram from KNX bus."""
        xknx = XKNX()
        test_device = DateTimeDevice(
            xknx, "TestDateTime", group_address="1/2/3", localtime=localtime
        )

        telegram_read = Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )
        with freeze_time("2017-07-11 09:13:14"):  # summer time
            test_device.process(telegram_read)

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(DPTArray(expected)),
        )

    async def test_process_read_custom_time(self) -> None:
        """Test test process a read telegram from KNX bus."""
        xknx = XKNX()
        test_device = TimeDevice(
            xknx,
            "TestDateTime",
            group_address="1/2/3",
            localtime=False,
            respond_to_read=True,
        )

        await test_device.set(dt.time(9, 13, 14))
        telegram_set = xknx.telegrams.get_nowait()
        assert telegram_set == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x9, 0xD, 0xE))),
        )
        test_device.process(telegram_set)

        telegram_read = Telegram(
            destination_address=GroupAddress("1/2/3"), payload=GroupValueRead()
        )
        test_device.process(telegram_read)

        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueResponse(DPTArray((0x9, 0xD, 0xE))),
        )

    #
    # TEST HAS GROUP ADDRESS
    #
    async def test_has_group_address_localtime(self) -> None:
        """Test if has_group_address function works."""
        xknx = XKNX()
        test_device = DateDevice(
            xknx,
            "TestDateTime",
            group_address="1/2/3",
            group_address_state="1/2/4",
            localtime=True,
        )
        assert test_device.has_group_address(GroupAddress("1/2/3"))
        # group_address_state ignored when using localtime
        assert not test_device.has_group_address(GroupAddress("1/2/4"))

    async def test_has_group_address_custom_time(self) -> None:
        """Test if has_group_address function works."""
        xknx = XKNX()
        test_device = DateDevice(
            xknx,
            "TestDateTime",
            group_address="1/2/3",
            group_address_state="1/2/4",
            localtime=False,
        )
        assert test_device.has_group_address(GroupAddress("1/2/3"))
        assert test_device.has_group_address(GroupAddress("1/2/4"))

    #
    # TEST BACKGROUND TASK
    #
    @patch("xknx.core.TelegramQueue.process_telegram_outgoing", new_callable=AsyncMock)
    async def test_background_task(
        self,
        process_telegram_outgoing_mock: AsyncMock,
        time_travel: EventLoopClockAdvancer,
        xknx_no_interface: XKNX,
    ) -> None:
        """Test if background task works."""
        xknx = xknx_no_interface
        xknx.connection_manager.connection_state_changed(XknxConnectionState.CONNECTED)
        test_device = TimeDevice(xknx, "TestDateTime", group_address="1/2/3")
        xknx.devices.async_add(test_device)
        async with xknx:
            # initial time telegram
            await time_travel(0)
            process_telegram_outgoing_mock.assert_called_once()
            process_telegram_outgoing_mock.reset_mock()
            # repeated time telegram
            await time_travel(BROADCAST_MINUTES * 60)
            process_telegram_outgoing_mock.assert_called_once()
            process_telegram_outgoing_mock.reset_mock()
            # remove device - no more telegrams
            xknx.devices.async_remove(test_device)
            await time_travel(BROADCAST_MINUTES * 60)
            process_telegram_outgoing_mock.assert_not_called()

    @patch("xknx.core.TelegramQueue.process_telegram_outgoing", new_callable=AsyncMock)
    async def test_no_background_task(
        self,
        process_telegram_outgoing_mock: AsyncMock,
        time_travel: EventLoopClockAdvancer,
        xknx_no_interface: XKNX,
    ) -> None:
        """Test if background task is not started when not using `localtime`."""
        xknx = xknx_no_interface
        test_device = TimeDevice(
            xknx,
            "TestDateTime",
            group_address="1/2/3",
            localtime=False,
        )
        xknx.devices.async_add(test_device)
        async with xknx:
            assert test_device._broadcast_task is None
            # no initial time telegram
            await time_travel(0)
            process_telegram_outgoing_mock.assert_not_called()
            # no repeated time telegram
            await time_travel(BROADCAST_MINUTES * 60)
            process_telegram_outgoing_mock.assert_not_called()



"""Unit test for KNX scene number."""

import pytest

from xknx.dpt import DPTArray, DPTSceneNumber
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestDPTSceneNumber:
    """Test class for KNX scaling value."""

    @pytest.mark.parametrize(
        ("raw", "value"),
        [
            ((0x31,), 50),
            ((0x3F,), 64),
            ((0x00,), 1),
        ],
    )
    def test_transcoder(self, raw: tuple[int], value: int) -> None:
        """Test parsing and streaming of DPTSceneNumber."""
        assert DPTSceneNumber.to_knx(value) == DPTArray(raw)
        assert DPTSceneNumber.from_knx(DPTArray(raw)) == value

    def test_to_knx_min_exceeded(self) -> None:
        """Test parsing of DPTSceneNumber with wrong value (underflow)."""
        with pytest.raises(ConversionError):
            DPTSceneNumber.to_knx(DPTSceneNumber.value_min - 1)

    def test_to_knx_max_exceeded(self) -> None:
        """Test parsing of DPTSceneNumber with wrong value (overflow)."""
        with pytest.raises(ConversionError):
            DPTSceneNumber.to_knx(DPTSceneNumber.value_max + 1)

    def test_to_knx_wrong_parameter(self) -> None:
        """Test parsing of DPTSceneNumber with wrong value (string)."""
        with pytest.raises(ConversionError):
            DPTSceneNumber.to_knx("fnord")

    def test_from_knx_wrong_parameter(self) -> None:
        """Test parsing of DPTSceneNumber with wrong value (3 byte array)."""
        with pytest.raises(CouldNotParseTelegram):
            DPTSceneNumber.from_knx(DPTArray((0x01, 0x02, 0x03)))

    def test_from_knx_wrong_value(self) -> None:
        """Test parsing of DPTSceneNumber with value which exceeds limits."""
        with pytest.raises(ConversionError):
            DPTSceneNumber.from_knx(DPTArray((0x64,)))



"""
Module for managing a cover via KNX.

It provides functionality for

* moving cover up/down or to a specific position
* reading the current state from KNX bus.
* Cover will also predict the current position.
"""

from __future__ import annotations

import asyncio
from collections.abc import Iterator
import logging
from typing import TYPE_CHECKING, Any, Final

from xknx.core import Task
from xknx.remote_value import (
    GroupAddressesType,
    RemoteValue,
    RemoteValueScaling,
    RemoteValueStep,
    RemoteValueSwitch,
    RemoteValueUpDown,
)

from .device import Device, DeviceCallbackType
from .travelcalculator import TravelCalculator, TravelStatus

if TYPE_CHECKING:
    from xknx.telegram import Telegram
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


DEFAULT_TRAVEL_TIME: Final = 25
TRAVELING_CALLBACK_INTERVAL: Final = 1


class Cover(Device):
    """Class for managing a cover."""

    def __init__(
        self,
        xknx: XKNX,
        name: str,
        group_address_long: GroupAddressesType = None,
        group_address_short: GroupAddressesType = None,
        group_address_stop: GroupAddressesType = None,
        group_address_position: GroupAddressesType = None,
        group_address_position_state: GroupAddressesType = None,
        group_address_angle: GroupAddressesType = None,
        group_address_angle_state: GroupAddressesType = None,
        group_address_locked_state: GroupAddressesType = None,
        sync_state: bool | int | float | str = True,
        travel_time_down: float = DEFAULT_TRAVEL_TIME,
        travel_time_up: float = DEFAULT_TRAVEL_TIME,
        invert_updown: bool = False,
        invert_position: bool = False,
        invert_angle: bool = False,
        device_updated_cb: DeviceCallbackType[Cover] | None = None,
    ) -> None:
        """Initialize Cover class."""
        super().__init__(xknx, name, device_updated_cb)
        # self.after_update for position changes is called after updating the
        # travelcalculator (in process_group_write and set_*) - angle changes
        # are updated from RemoteValue objects
        self.updown = RemoteValueUpDown(
            xknx,
            group_address_long,
            device_name=self.name,
            after_update_cb=None,
            invert=invert_updown,
        )

        self.step = RemoteValueStep(
            xknx,
            group_address_short,
            device_name=self.name,
            after_update_cb=self.after_update,
            invert=invert_updown,
        )

        self.stop_ = RemoteValueSwitch(
            xknx,
            group_address=group_address_stop,
            sync_state=False,
            device_name=self.name,
            after_update_cb=None,
        )

        position_range_from = 100 if invert_position else 0
        position_range_to = 0 if invert_position else 100
        self.position_current = RemoteValueScaling(
            xknx,
            group_address_state=group_address_position_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Position",
            after_update_cb=self._current_position_from_rv,
            range_from=position_range_from,
            range_to=position_range_to,
        )
        self.position_target = RemoteValueScaling(
            xknx,
            group_address=group_address_position,
            device_name=self.name,
            feature_name="Target position",
            after_update_cb=self._target_position_from_rv,
            range_from=position_range_from,
            range_to=position_range_to,
        )

        angle_range_from = 100 if invert_angle else 0
        angle_range_to = 0 if invert_angle else 100
        self.angle = RemoteValueScaling(
            xknx,
            group_address_angle,
            group_address_angle_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Tilt angle",
            after_update_cb=self.after_update,
            range_from=angle_range_from,
            range_to=angle_range_to,
        )

        self.locked = RemoteValueSwitch(
            xknx,
            group_address_state=group_address_locked_state,
            sync_state=sync_state,
            device_name=self.name,
            feature_name="Locked",
            after_update_cb=self.after_update,
        )

        self.travel_time_down = travel_time_down
        self.travel_time_up = travel_time_up
        self.travelcalculator = TravelCalculator(travel_time_down, travel_time_up)

        self._auto_stop_task = Task(
            name=f"cover.auto_stopper_{id(self)}",
            target=self.stop,
        )
        self._auto_stop_requested: bool = False
        self._periodic_update_task = Task(
            name=f"cover.periodic_update_{id(self)}",
            target=self._periodic_updater,
        )
        self._travel_direction_tilt: TravelStatus | None = None

    def _iter_remote_values(self) -> Iterator[RemoteValue[Any]]:
        """Iterate the devices RemoteValue classes."""
        yield self.updown
        yield self.step
        yield self.stop_
        yield self.position_current
        yield self.position_target
        yield self.angle
        yield self.locked

    def async_remove_tasks(self) -> None:
        """Remove async tasks of device."""
        self.xknx.task_registry.remove_task(self._auto_stop_task)
        self._auto_stop_requested = False
        self.xknx.task_registry.remove_task(self._periodic_update_task)

    async def set_down(self) -> None:
        """Move cover down."""
        if self.updown.writable:
            self.updown.down()
            self._travel_direction_tilt = None
            self._start_position_update(
                target_position=self.travelcalculator.position_closed
            )
        elif self.position_target.writable:
            self.position_target.set(self.travelcalculator.position_closed)

    async def set_up(self) -> None:
        """Move cover up."""
        if self.updown.writable:
            self.updown.up()
            self._travel_direction_tilt = None
            self._start_position_update(
                target_position=self.travelcalculator.position_open
            )
        elif self.position_target.writable:
            self.position_target.set(self.travelcalculator.position_open)

    async def set_short_down(self) -> None:
        """Move cover short down."""
        self.step.increase()

    async def set_short_up(self) -> None:
        """Move cover short up."""
        self.step.decrease()

    async def stop(self) -> None:
        """Stop cover."""
        if self.stop_.writable:
            self.stop_.on()
        elif self.step.writable:
            if TravelStatus.DIRECTION_UP in (
                self.travelcalculator.travel_direction,
                self._travel_direction_tilt,
            ):
                self.step.decrease()
            elif TravelStatus.DIRECTION_DOWN in (
                self.travelcalculator.travel_direction,
                self._travel_direction_tilt,
            ):
                self.step.increase()
        else:
            logger.warning("Stop not supported for device %s", self.get_name())
            return
        self._travel_direction_tilt = None
        self._stop_position_update()

    async def set_position(self, position: int) -> None:
        """Move cover to a desginated position."""
        if self.position_target.writable:
            self.position_target.set(position)
            return
        # No direct positioning group address defined
        # fully open or close is always possible even if current position is not known
        current_position = self.travelcalculator.current_position()
        if current_position is None:
            if position == self.travelcalculator.position_open:
                self.updown.up()
            elif position == self.travelcalculator.position_closed:
                self.updown.down()
            else:
                logger.warning(
                    "Current position unknown. Initialize cover by moving to end position."
                )
                return
            self._start_position_update(target_position=position)
            return
        if position < current_position:
            self.updown.up()
        elif position > current_position:
            self.updown.down()
        else:
            return  # already in position
        self._start_position_update(target_position=position)
        if self.supports_stop:
            # If device does not support positioning, we stop the device when position is reached
            self._start_auto_stopper(current_position, position)

    def _start_position_update(self, target_position: int) -> None:
        """Start the travel calculator and run device callbacks."""
        self.travelcalculator.start_travel(target_position)
        self.after_update()
        if self.travelcalculator.is_traveling():
            # restarts when already running
            self.xknx.task_registry.start_task(self._periodic_update_task)

    async def _periodic_updater(self) -> None:
        """Run callback periodically while traveling."""
        while self.travelcalculator.is_traveling():
            await asyncio.sleep(TRAVELING_CALLBACK_INTERVAL)
            if self.travelcalculator.is_traveling():
                # else _stop_position_update will call after_update a second time
                self.after_update()
        self._stop_position_update()

    def _stop_position_update(self) -> None:
        """Stop the travel calculator and periodic device callbacks."""
        if not self.travelcalculator.position_reached():
            self.travelcalculator.stop()
        self._periodic_update_task.cancel()
        self.after_update()

    def _start_auto_stopper(self, current_position: int, target_position: int) -> None:
        """Start or restart the auto stopper task."""
        # Unless traveling to fully open or fully closed state, send stop command
        if target_position in (
            self.travelcalculator.position_open,
            self.travelcalculator.position_closed,
        ):
            return

        stop_in_seconds = self.travelcalculator.calculate_travel_time(
            from_position=current_position, to_position=target_position
        )
        self._auto_stop_task.wait_before_start = stop_in_seconds
        self.xknx.task_registry.start_task(self._auto_stop_task)
        self._auto_stop_requested = True

    def _cancel_auto_stopper(self) -> None:
        """Cancel the auto stopper task."""
        self._auto_stop_task.cancel()
        self._auto_stop_requested = False

    def _target_position_from_rv(self, new_target_postion: int) -> None:
        """Update the target position from RemoteValue (Callback)."""
        self._start_position_update(target_position=new_target_postion)

    def _current_position_from_rv(self, new_position: int) -> None:
        """Update the current position from RemoteValue (Callback)."""
        position_before_update = self.travelcalculator.current_position()
        if self.is_traveling():
            self.travelcalculator.update_position(new_position)
        else:
            self.travelcalculator.set_position(new_position)
        if position_before_update != self.travelcalculator.current_position():
            if position_before_update is not None:  # None on first move
                # restarts when already running
                self.xknx.task_registry.start_task(self._periodic_update_task)
            self.after_update()

    async def set_angle(self, angle: int) -> None:
        """Move cover to designated angle."""
        if not self.supports_angle:
            logger.warning("Angle not supported for device %s", self.get_name())
            return

        current_angle = self.current_angle()
        self._travel_direction_tilt = (
            TravelStatus.DIRECTION_DOWN
            if current_angle is not None and angle >= current_angle
            else TravelStatus.DIRECTION_UP
        )

        self.angle.set(angle)

    async def sync(self, wait_for_result: bool = False) -> None:
        """Read states of device from KNX bus."""
        await self.position_current.read_state(wait_for_result=wait_for_result)
        await self.angle.read_state(wait_for_result=wait_for_result)

    def process_group_write(self, telegram: Telegram) -> None:
        """Process incoming and outgoing GROUP WRITE telegram."""
        # call after_update to account for travelcalculator changes
        if self.updown.process(telegram):
            if self._auto_stop_requested:
                # Don't cancel auto stopper if we initiated the up/down telegram for it
                self._auto_stop_requested = False
            else:
                self._cancel_auto_stopper()
            if (
                not self.is_opening()
                and self.updown.value == RemoteValueUpDown.Direction.UP
            ):
                self._start_position_update(
                    target_position=self.travelcalculator.position_open
                )
            elif (
                not self.is_closing()
                and self.updown.value == RemoteValueUpDown.Direction.DOWN
            ):
                self._start_position_update(
                    target_position=self.travelcalculator.position_closed
                )
        # stop from bus
        if self.stop_.process(telegram) or self.step.process(telegram):
            self._cancel_auto_stopper()
            if self.is_traveling():
                self._stop_position_update()

        self.position_current.process(telegram, always_callback=True)
        self.position_target.process(telegram, always_callback=True)
        self.angle.process(telegram)
        self.locked.process(telegram)

    def current_position(self) -> int | None:
        """Return current position of cover."""
        return self.travelcalculator.current_position()

    def current_angle(self) -> int | None:
        """Return current tilt angle of cover."""
        return self.angle.value

    def is_locked(self) -> bool | None:
        """Return if the cover is currently locked for manual movement."""
        return self.locked.value

    def is_traveling(self) -> bool:
        """Return if cover is traveling at the moment."""
        return self.travelcalculator.is_traveling()

    def position_reached(self) -> bool:
        """Return if cover has reached its final position."""
        return self.travelcalculator.position_reached()

    def is_open(self) -> bool:
        """Return if cover is open."""
        return self.travelcalculator.is_open()

    def is_closed(self) -> bool:
        """Return if cover is closed."""
        return self.travelcalculator.is_closed()

    def is_opening(self) -> bool:
        """Return if the cover is opening or not."""
        return self.travelcalculator.is_opening()

    def is_closing(self) -> bool:
        """Return if the cover is closing or not."""
        return self.travelcalculator.is_closing()

    @property
    def supports_stop(self) -> bool:
        """Return if cover supports manual stopping."""
        return self.stop_.writable or self.step.writable

    @property
    def supports_locked(self) -> bool:
        """Return if cover supports locking."""
        return self.locked.initialized

    @property
    def supports_position(self) -> bool:
        """Return if cover supports direct positioning."""
        return self.position_target.initialized

    @property
    def supports_angle(self) -> bool:
        """Return if cover supports tilt angle."""
        return self.angle.initialized

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            f'<Cover name="{self.name}" '
            f"updown={self.updown.group_addr_str()} "
            f"step={self.step.group_addr_str()} "
            f"stop_={self.stop_.group_addr_str()} "
            f"position_current={self.position_current.group_addr_str()} "
            f"position_target={self.position_target.group_addr_str()} "
            f"angle={self.angle.group_addr_str()} "
            f"locked={self.locked.group_addr_str()} "
            f'travel_time_down="{self.travel_time_down}" '
            f'travel_time_up="{self.travel_time_up}" '
            "/>"
        )



"""Example for reading the serial number of a KNX device."""

import asyncio
import sys

from xknx import XKNX
from xknx.management import procedures
from xknx.profile import ResourceGenericPropertyId
from xknx.telegram import IndividualAddress, apci


async def read_serial_number(
    xknx: XKNX, individual_address: IndividualAddress
) -> bytes | None:
    """Read the serial number of a KNX device."""
    async with xknx.management.connection(address=individual_address) as connection:
        response = await connection.request(
            payload=apci.PropertyValueRead(
                property_id=ResourceGenericPropertyId.PID_SERIAL_NUMBER
            ),
            expected=apci.PropertyValueResponse,
        )
    if response.payload is None:
        print(f"Reading serial number from {individual_address} failed.")
        return None

    assert isinstance(response.payload, apci.PropertyValueResponse)
    return response.payload.data


async def main(argv: list[str]) -> int:
    """
    Read serial number from a device by given individual address or in programming mode.

    When no individual address is given, a device in programming mode is searched for.
    This fails if multiple devices are in programming mode and/or when there is no device found in programming mode.
    """
    individual_address: IndividualAddress | None = None

    if len(argv) == 2:
        individual_address = IndividualAddress(argv[1])
        print(
            f"Using individual address {individual_address} from command line argument."
        )

    async with XKNX() as xknx:
        if not individual_address:
            print("Searching for device in programming mode...")
            found = await procedures.nm_individual_address_read(
                xknx, raise_if_multiple=True
            )
            if not found:
                print("No device in programming mode found.")
                return 1
            individual_address = found[0]
            print(
                f"Found device in programming mode with individual address {individual_address}."
            )

        serial_number = await read_serial_number(xknx, individual_address)
        if serial_number is None:
            print("Failed to read serial number.")
            return 1
        # format the serial number in same style as ETS
        serial_string = f"{serial_number[:2].hex()}:{serial_number[2:].hex()}"
        print(f"Serial number of {individual_address} is {serial_string}.")

    return 0


if __name__ == "__main__":
    asyncio.run(main(sys.argv))



"""Unit test for KNX time objects."""

import datetime
from typing import Any

import pytest

from xknx.dpt import DPTArray, DPTBinary, DPTTime
from xknx.dpt.dpt_10 import KNXDay, KNXTime
from xknx.exceptions import ConversionError, CouldNotParseTelegram


class TestKNXTime:
    """Test class for KNX time objects."""

    @pytest.mark.parametrize(
        ("data", "value"),
        [
            (
                {"hour": 5, "minutes": 10, "seconds": 9},
                KNXTime(5, 10, 9),
            ),
            (
                {"hour": 21, "minutes": 52, "seconds": 9, "day": "tuesday"},
                KNXTime(21, 52, 9, KNXDay.TUESDAY),
            ),
            (
                {"hour": 0, "minutes": 0, "seconds": 0},
                KNXTime(0, 0, 0, KNXDay.NO_DAY),
            ),
        ],
    )
    def test_dict(self, data: dict[str, Any], value: KNXTime) -> None:
        """Test from_dict and as_dict methods."""
        assert KNXTime.from_dict(data) == value
        # day defaults to `no_day`
        default_dict = {"day": "no_day"}
        assert value.as_dict() == default_dict | data

    @pytest.mark.parametrize(
        "data",
        [
            # invalid data
            {"hour": 1},
            {"hour": "a"},
            {"minutes": 2, "seconds": 3},
            {"hour": 2, "seconds": 3},
            {"hour": 1, "minutes": 2},
            {"hour": 1, "minutes": 2, "seconds": "a"},
            {"hour": 1, "minutes": 2, "seconds": 3, "day": "a"},
        ],
    )
    def test_dict_invalid(self, data: dict[str, Any]) -> None:
        """Test from_dict and as_dict methods."""
        with pytest.raises(ValueError):
            KNXTime.from_dict(data)

    @pytest.mark.parametrize(
        ("time", "value"),
        [
            (datetime.time(5, 10, 9), KNXTime(5, 10, 9)),
            (datetime.time(21, 52, 9), KNXTime(21, 52, 9)),
            (datetime.time(0, 0, 0), KNXTime(0, 0, 0)),
            (datetime.time(23, 59, 59), KNXTime(23, 59, 59)),
        ],
    )
    def test_as_time(self, time: datetime.time, value: KNXTime) -> None:
        """Test from_time and as_time methods."""
        assert KNXTime.from_time(time) == value
        assert value.as_time() == time


class TestDPTTime:
    """Test class for KNX time objects."""

    @pytest.mark.parametrize(
        ("value", "raw"),
        [
            (KNXTime(13, 23, 42, KNXDay.TUESDAY), (0x4D, 0x17, 0x2A)),
            (KNXTime(0, 0, 0, KNXDay.NO_DAY), (0x0, 0x0, 0x0)),
            (KNXTime(23, 59, 59, KNXDay.SUNDAY), (0xF7, 0x3B, 0x3B)),
        ],
    )
    def test_parse(self, value: KNXTime, raw: tuple[int, ...]) -> None:
        """Test parsing and streaming."""
        knx_value = DPTTime.to_knx(value)
        assert knx_value == DPTArray(raw)
        assert DPTTime.from_knx(knx_value) == value

    def test_from_knx_wrong_value(self) -> None:
        """Test parsing from DPTTime object from wrong binary values."""
        with pytest.raises(ConversionError):
            # this parameter exceeds limit
            DPTTime.from_knx(DPTArray((0xF7, 0x3B, 0x3C)))
        with pytest.raises(CouldNotParseTelegram):
            DPTTime.from_knx(DPTArray((0xFF, 0x4E)))
        with pytest.raises(CouldNotParseTelegram):
            DPTTime.from_knx(DPTBinary(True))

    def test_to_knx_wrong_parameter(self) -> None:
        """Test parsing from DPTTime object from wrong value."""
        with pytest.raises(ConversionError):
            DPTTime.to_knx(KNXTime(24, 0, 0))  # out of range
        with pytest.raises(ConversionError):
            DPTTime.to_knx(KNXTime(0, 60, 0))  # out of range
        with pytest.raises(ConversionError):
            DPTTime.to_knx("fnord")
        with pytest.raises(ConversionError):
            DPTTime.to_knx((1, 2, 3))



"""Unit test for devices container within XKNX."""

from unittest.mock import AsyncMock, Mock, patch

import pytest

from xknx import XKNX
from xknx.devices import BinarySensor, Device, Light, Switch
from xknx.telegram import GroupAddress


class TestDevices:
    """Test class for devices container within XKNX."""

    #
    # XKNX Config
    #
    def test_get_item(self) -> None:
        """Test get item by name or by index."""
        xknx = XKNX()
        light1 = Light(xknx, "Living-Room.Light_1", group_address_switch="1/6/7")
        xknx.devices.async_add(light1)
        switch1 = Switch(xknx, "TestOutlet_1", group_address="1/2/3")
        xknx.devices.async_add(switch1)
        light2 = Light(xknx, "Living-Room.Light_2", group_address_switch="1/6/8")
        xknx.devices.async_add(light2)
        switch2 = Switch(xknx, "TestOutlet_2", group_address="1/2/4")
        xknx.devices.async_add(switch2)

        assert xknx.devices["Living-Room.Light_1"] == light1
        assert xknx.devices["TestOutlet_1"] == switch1
        assert xknx.devices["Living-Room.Light_2"] == light2
        assert xknx.devices["TestOutlet_2"] == switch2
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            xknx.devices["TestOutlet_2sdds"]

        assert xknx.devices[0] == light1
        assert xknx.devices[1] == switch1
        assert xknx.devices[2] == light2
        assert xknx.devices[3] == switch2
        with pytest.raises(IndexError):
            # pylint: disable=pointless-statement
            xknx.devices[4]

    def test_device_by_group_address(self) -> None:
        """Test get devices by group address."""
        xknx = XKNX()

        light1 = Light(xknx, "Livingroom", group_address_switch="1/6/7")
        sensor1 = BinarySensor(xknx, "Diningroom", group_address_state="3/0/1")
        sensor2 = BinarySensor(xknx, "Diningroom", group_address_state="3/0/1")
        light2 = Light(xknx, "Livingroom", group_address_switch="1/6/8")
        xknx.devices.async_add(light1)
        xknx.devices.async_add(sensor1)
        xknx.devices.async_add(sensor2)
        xknx.devices.async_add(light2)

        assert tuple(xknx.devices.devices_by_group_address(GroupAddress("1/6/7"))) == (
            light1,
        )
        assert tuple(xknx.devices.devices_by_group_address(GroupAddress("1/6/8"))) == (
            light2,
        )
        assert tuple(xknx.devices.devices_by_group_address(GroupAddress("3/0/1"))) == (
            sensor1,
            sensor2,
        )

    def test_iter(self) -> None:
        """Test __iter__() function."""
        xknx = XKNX()

        light1 = Light(xknx, "Livingroom", group_address_switch="1/6/7")
        sensor1 = BinarySensor(xknx, "Diningroom", group_address_state="3/0/1")
        sensor2 = BinarySensor(xknx, "Diningroom", group_address_state="3/0/1")
        light2 = Light(xknx, "Livingroom", group_address_switch="1/6/8")
        xknx.devices.async_add(light1)
        xknx.devices.async_add(sensor1)
        xknx.devices.async_add(sensor2)
        xknx.devices.async_add(light2)

        assert tuple(iter(xknx.devices)) == (light1, sensor1, sensor2, light2)

    def test_len(self) -> None:
        """Test len() function."""
        xknx = XKNX()
        assert len(xknx.devices) == 0

        light = Light(xknx, "Living-Room.Light_1", group_address_switch="1/6/7")
        xknx.devices.async_add(light)
        assert len(xknx.devices) == 1

        binary_sensor = BinarySensor(
            xknx, "DiningRoom.Motion.Sensor", group_address_state="3/0/1"
        )
        xknx.devices.async_add(binary_sensor)
        assert len(xknx.devices) == 2

        xknx.devices.async_remove(light)
        assert len(xknx.devices) == 1

        xknx.devices.async_add(light)
        assert len(xknx.devices) == 2

    def test_contains(self) -> None:
        """Test __contains__() function."""
        xknx = XKNX()
        xknx.devices.async_add(
            Light(xknx, "Living-Room.Light_1", group_address_switch="1/6/7")
        )
        xknx.devices.async_add(
            Light(xknx, "Living-Room.Light_2", group_address_switch="1/6/8")
        )

        assert "Living-Room.Light_1" in xknx.devices
        assert "Living-Room.Light_2" in xknx.devices
        assert "Living-Room.Light_3" not in xknx.devices

    @patch.multiple(Device, __abstractmethods__=set())
    def test_add_remove(self) -> None:
        """Tesst add and remove functions."""
        xknx = XKNX()
        device1 = Device(xknx, "TestDevice1")
        device2 = Device(xknx, "TestDevice2")
        xknx.devices.async_add(device1)
        xknx.devices.async_add(device2)
        assert len(xknx.devices) == 2
        xknx.devices.async_remove(device1)
        assert len(xknx.devices) == 1
        assert "TestDevice1" not in xknx.devices
        xknx.devices.async_remove(device2)
        assert len(xknx.devices) == 0

    async def test_modification_of_device(self) -> None:
        """Test if devices object does store references and not copies of objects."""
        xknx = XKNX()
        light1 = Light(xknx, "Living-Room.Light_1", group_address_switch="1/6/7")
        xknx.devices.async_add(light1)
        for device in xknx.devices:
            await device.set_on()
            xknx.devices.process(xknx.telegrams.get_nowait())
        assert light1.state

        device2 = xknx.devices["Living-Room.Light_1"]
        await device2.set_off()
        xknx.devices.process(xknx.telegrams.get_nowait())
        assert not light1.state

        for device in xknx.devices.devices_by_group_address(GroupAddress("1/6/7")):
            await device.set_on()
            xknx.devices.process(xknx.telegrams.get_nowait())
        assert light1.state

    #
    # TEST SYNC
    #
    @patch.multiple(Device, __abstractmethods__=set())
    async def test_sync(self) -> None:
        """Test sync function."""
        xknx = XKNX()
        xknx.devices.async_add(Device(xknx, "TestDevice1"))
        xknx.devices.async_add(Device(xknx, "TestDevice2"))
        with patch("xknx.devices.Device.sync", new_callable=AsyncMock) as mock_sync:
            await xknx.devices.sync()
            assert mock_sync.call_count == 2

    #
    # TEST CALLBACK
    #
    @patch.multiple(Device, __abstractmethods__=set())
    def test_device_updated_callback(self) -> None:
        """Test if device updated callback is called correctly if device was updated."""
        xknx = XKNX()
        device1 = Device(xknx, "TestDevice1")
        device2 = Device(xknx, "TestDevice2")
        xknx.devices.async_add(device1)
        xknx.devices.async_add(device2)

        after_update_callback1 = Mock()
        after_update_callback2 = Mock()

        # Registering both callbacks
        xknx.devices.register_device_updated_cb(after_update_callback1)
        xknx.devices.register_device_updated_cb(after_update_callback2)

        # Triggering first device. Both callbacks to be called
        device1.after_update()
        after_update_callback1.assert_called_with(device1)
        after_update_callback2.assert_called_with(device1)
        after_update_callback1.reset_mock()
        after_update_callback2.reset_mock()

        # Triggering 2nd device. Both callbacks have to be called
        device2.after_update()
        after_update_callback1.assert_called_with(device2)
        after_update_callback2.assert_called_with(device2)
        after_update_callback1.reset_mock()
        after_update_callback2.reset_mock()

        # Unregistering first callback
        xknx.devices.unregister_device_updated_cb(after_update_callback1)

        # Triggering first device. Only second callback has to be called
        device1.after_update()
        after_update_callback1.assert_not_called()
        after_update_callback2.assert_called_with(device1)
        after_update_callback1.reset_mock()
        after_update_callback2.reset_mock()

        # Unregistering second callback
        xknx.devices.unregister_device_updated_cb(after_update_callback2)

        # Triggering second device. No callback should be called
        device2.after_update()
        after_update_callback1.assert_not_called()
        after_update_callback2.assert_not_called()
        after_update_callback1.reset_mock()
        after_update_callback2.reset_mock()



"""Example for switching a light on and off."""

import asyncio

from xknx import XKNX
from xknx.devices import Light


async def main() -> None:
    """Connect to KNX/IP bus, switch on light, wait 2 seconds and switch of off again."""
    xknx = XKNX()
    await xknx.start()
    light = Light(xknx, name="TestLight", group_address_switch="1/0/9")
    xknx.devices.async_add(light)

    await light.set_on()
    await asyncio.sleep(2)
    await light.set_off()
    await xknx.stop()


asyncio.run(main())



"""Unit test for RemoteValueColorRGBW objects."""

import pytest

from xknx import XKNX
from xknx.dpt import DPTArray, DPTBinary, RGBWColor
from xknx.exceptions import ConversionError
from xknx.remote_value import RemoteValueColorRGBW
from xknx.telegram import GroupAddress, Telegram
from xknx.telegram.apci import GroupValueWrite


class TestRemoteValueColorRGBW:
    """Test class for RemoteValueColorRGBW objects."""

    def test_to_knx(self) -> None:
        """Test to_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGBW(xknx)
        input_value = RGBWColor(100, 101, 102, 127)
        expected = DPTArray((0x64, 0x65, 0x66, 0x7F, 0x00, 0x0F))
        assert remote_value.to_knx(input_value) == expected

    def test_from_knx(self) -> None:
        """Test from_knx function with normal operation."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGBW(xknx)
        assert remote_value.from_knx(
            DPTArray((0x64, 0x65, 0x66, 0x7F, 0x00, 0x00))
        ) == RGBWColor(None, None, None, None)
        assert remote_value.from_knx(
            DPTArray((0x64, 0x65, 0x66, 0x7F, 0x00, 0x0F))
        ) == RGBWColor(100, 101, 102, 127)
        assert remote_value.from_knx(
            DPTArray((0x64, 0x65, 0x66, 0x7F, 0x00, 0x00))
        ) == RGBWColor(100, 101, 102, 127)
        assert remote_value.from_knx(
            DPTArray((0xFF, 0x65, 0x66, 0xFF, 0x00, 0x09))
        ) == RGBWColor(255, 101, 102, 255)
        assert remote_value.from_knx(
            DPTArray((0x64, 0x65, 0x66, 0x7F, 0x00, 0x01))
        ) == RGBWColor(255, 101, 102, 127)

    def test_to_knx_error(self) -> None:
        """Test to_knx function with wrong parametern."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGBW(xknx)
        with pytest.raises(ConversionError):
            remote_value.to_knx((101, 102, 103, 104))

    async def test_set(self) -> None:
        """Test setting value."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGBW(xknx, group_address=GroupAddress("1/2/3"))
        remote_value.set(RGBWColor(100, 101, 102, 103))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x66, 0x67, 0x00, 0x0F))),
        )
        remote_value.set(RGBWColor(100, 101, 104, 105))
        assert xknx.telegrams.qsize() == 1
        telegram = xknx.telegrams.get_nowait()
        assert telegram == Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x68, 0x69, 0x00, 0x0F))),
        )

    def test_process(self) -> None:
        """Test process telegram."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGBW(xknx, group_address=GroupAddress("1/2/3"))
        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x66, 0x67, 0x00, 0x0F))),
        )
        remote_value.process(telegram)
        assert remote_value.value == RGBWColor(100, 101, 102, 103)

    def test_to_process_error(self) -> None:
        """Test process erroneous telegram."""
        xknx = XKNX()
        remote_value = RemoteValueColorRGBW(xknx, group_address=GroupAddress("1/2/3"))

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTBinary(1)),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(DPTArray((0x64, 0x65, 0x66))),
        )
        assert remote_value.process(telegram) is False

        telegram = Telegram(
            destination_address=GroupAddress("1/2/3"),
            payload=GroupValueWrite(
                DPTArray((0x00, 0x00, 0x0F, 0x64, 0x65, 0x66, 0x67))
            ),
        )
        assert remote_value.process(telegram) is False

        assert remote_value.value is None



"""
Abstraction for handling KNX/IP tunnels.

Tunnels connect to KNX/IP devices directly via UDP or TCP and build a static connection.
"""

from __future__ import annotations

from abc import abstractmethod
import asyncio
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
import logging
from typing import TYPE_CHECKING

from xknx.cemi import CEMIFrame
from xknx.core import XknxConnectionState, XknxConnectionType
from xknx.exceptions import CommunicationError, TunnellingAckError
from xknx.knxip import (
    HPAI,
    ConnectRequestInformation,
    DisconnectRequest,
    DisconnectResponse,
    HostProtocol,
    KNXIPFrame,
    KNXIPServiceType,
    TunnellingAck,
    TunnellingRequest,
)
from xknx.telegram import IndividualAddress

from .const import HEARTBEAT_RATE
from .gateway_scanner import GatewayDescriptor
from .interface import CEMIBytesCallbackType, Interface
from .ip_secure import SecureSession
from .request_response import Connect, ConnectionState, Disconnect, Tunnelling
from .self_description import DescriptionQuery
from .transport import KNXIPTransport, TCPTransport, UDPTransport

if TYPE_CHECKING:
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


class _Tunnel(Interface):
    """Class for handling KNX/IP tunnels."""

    __slots__ = (
        "_data_endpoint_addr",
        "_heartbeat_task",
        "_reconnect_task",
        "_requested_address",
        "_send_lock",
        "_src_address",
        "auto_reconnect",
        "auto_reconnect_wait",
        "cemi_received_callback",
        "communication_channel",
        "local_hpai",
        "sequence_number",
        "xknx",
    )

    connection_type: XknxConnectionType
    transport: KNXIPTransport

    def __init__(
        self,
        xknx: XKNX,
        cemi_received_callback: CEMIBytesCallbackType,
        auto_reconnect: bool = True,
        auto_reconnect_wait: int = 3,
    ) -> None:
        """Initialize Tunnel class."""
        self.xknx = xknx
        self.auto_reconnect = auto_reconnect
        self.auto_reconnect_wait = auto_reconnect_wait

        self.communication_channel: int | None = None
        self.local_hpai: HPAI = HPAI()
        self.sequence_number = 0
        self.cemi_received_callback = cemi_received_callback
        self._data_endpoint_addr: tuple[str, int] | None = None
        self._heartbeat_task: asyncio.Task[None] | None = None
        self._reconnect_task: asyncio.Task[None] | None = None
        self._requested_address: IndividualAddress | None = None
        self._src_address = IndividualAddress(0)
        self._send_lock = asyncio.Lock()

        self._init_transport()
        self.transport.register_callback(
            self._request_received,
            [
                KNXIPServiceType.TUNNELLING_REQUEST,
                KNXIPServiceType.DISCONNECT_REQUEST,
            ],
        )

    @abstractmethod
    def _init_transport(self) -> None:
        """Initialize transport."""
        # set up self.transport

    @abstractmethod
    async def setup_tunnel(self) -> None:
        """Set up tunnel before sending a ConnectionRequest."""
        # eg. set local HPAI used for control and data endpoint

    ####################
    #
    # CONNECT DISCONNECT
    #
    ####################

    async def connect(self) -> None:
        """
        Connect to a KNX tunneling interface.

        Raise CommunicationError when not successful.
        """
        self.xknx.connection_manager.connection_state_changed(
            XknxConnectionState.CONNECTING, self.connection_type
        )
        try:
            await self.transport.connect()
            await self.setup_tunnel()
            await self._connect_request()
        except (OSError, CommunicationError) as ex:
            logger.debug(
                "Could not establish connection to KNX/IP interface. %s: %s",
                type(ex).__name__,
                ex,
            )
            self.xknx.connection_manager.connection_state_changed(
                XknxConnectionState.DISCONNECTED
            )
            # close transport to prevent open file descriptors
            self.transport.stop()
            raise CommunicationError(
                "Tunnel connection could not be established"
            ) from ex

        self._tunnel_established()
        self.xknx.connection_manager.connection_state_changed(
            XknxConnectionState.CONNECTED, self.connection_type
        )

    def _tunnel_established(self) -> None:
        """Set up interface when the tunnel is ready."""
        self.sequence_number = 0
        self.start_heartbeat()

    def _tunnel_lost(self) -> None:
        """Prepare for reconnection or shutdown when the connection is lost. Callback."""
        if self.auto_reconnect:
            # _tunnel_lost might be called multiple times when the transport receives
            # multiple invalid frames - ensure only one reconnect task is started
            if self._reconnect_task is None:

                def _reconnect_task_cleanup(_: asyncio.Task[None]) -> None:
                    """Cleanup task so we don't need to check `done()` or do it explicitly everywhere."""
                    self._reconnect_task = None

                self._reconnect_task = asyncio.create_task(self._reconnect())
                self._reconnect_task.add_done_callback(_reconnect_task_cleanup)
            else:
                logger.debug("Reconnect already in progress.")
            return

        # no reconnect - clean up, close transport, raise
        self._prepare_disconnect()
        if self.transport.transport:
            if self.communication_channel is not None:
                # Don't wait for DisconnectResponse (no async here anyway)
                disconnect_request = DisconnectRequest(
                    communication_channel_id=self.communication_channel,
                    control_endpoint=self.local_hpai,
                )
                self.transport.send(KNXIPFrame.init_from_body(disconnect_request))
            self.transport.stop()
        logger.warning("Tunnel connection closed. auto_reconnect is disabled.")

    async def _reconnect(self) -> None:
        """Reconnect to tunnel device."""
        self._prepare_disconnect()
        if self.transport.transport:
            # when server issued DisconnectRequest communication_channel is already None so this is a no-op
            await self._disconnect_request()
            self._data_endpoint_addr = None
            self.transport.stop()

        attempt = 1
        while True:
            try:
                logger.debug("Reconnecting to KNX bus... (attempt %s)", attempt)
                await self.connect()
            except CommunicationError:
                attempt += 1
                logger.debug(
                    "Reconnection to KNX bus failed. Trying again in %s seconds.",
                    self.auto_reconnect_wait,
                )
                await asyncio.sleep(self.auto_reconnect_wait)
            else:
                logger.info("Successfully reconnected to KNX bus.")
                break

    def _prepare_disconnect(self) -> None:
        """Prepare for disconnect. Stop tunnel related tasks and set connection state."""
        self.stop_heartbeat()
        self.xknx.connection_manager.connection_state_changed(
            XknxConnectionState.DISCONNECTED
        )

    def _stop_reconnect(self) -> None:
        """Stop reconnect task if running."""
        if self._reconnect_task is not None:
            self._reconnect_task.cancel()

    async def disconnect(self) -> None:
        """Disconnect tunneling connection."""
        self._prepare_disconnect()
        self._stop_reconnect()
        try:
            await self._disconnect_request()
        finally:
            self.transport.stop()

    ####################
    #
    # OUTGOING REQUESTS
    #
    ####################

    async def _connect_request(self) -> bool:
        """Connect to tunnelling server. Set communication_channel and src_address."""
        connect = Connect(
            transport=self.transport,
            local_hpai=self.local_hpai,
            cri=ConnectRequestInformation(individual_address=self._requested_address),
        )
        await connect.start()
        if connect.success:
            self.communication_channel = connect.communication_channel
            # assign data_endpoint received from server
            self._data_endpoint_addr = (
                None
                if connect.data_endpoint.route_back
                else connect.data_endpoint.addr_tuple
            )
            # Use the individual address provided by the tunnelling server
            self._src_address = connect.crd.individual_address or IndividualAddress(0)
            self.xknx.current_address = self._src_address
            logger.debug(
                "Tunnel established. communication_channel=%s, address=%s",
                connect.communication_channel,
                self._src_address,
            )
            return True
        raise CommunicationError(
            f"ConnectRequest failed. Status code: {connect.response_status_code}"
        )

    async def _connectionstate_request(self) -> tuple[bool, str | None]:
        """Return state of tunnel. True if tunnel is in good shape."""
        if self.communication_channel is None:
            raise CommunicationError("No active communication channel.")
        conn_state = ConnectionState(
            transport=self.transport,
            communication_channel_id=self.communication_channel,
            local_hpai=self.local_hpai,
        )
        await conn_state.start()

        status_code: str | None = None
        if error_code := conn_state.response_status_code:
            status_code = error_code.name
        return conn_state.success, status_code

    async def _disconnect_request(self) -> None:
        """Disconnect from tunnel device. Delete communication_channel."""
        if self.communication_channel is not None:
            disconnect = Disconnect(
                transport=self.transport,
                communication_channel_id=self.communication_channel,
                local_hpai=self.local_hpai,
            )
            await disconnect.start()
            if disconnect.success:
                logger.debug(
                    "Tunnel disconnect succeeded (communication_channel: %s)",
                    self.communication_channel,
                )
            else:
                error_msg = (
                    f"with error code: {disconnect.response_status_code}"
                    if disconnect.response_status_code is not None
                    else "- no response from device (timeout)"
                )
                logger.warning(
                    "Tunnel disconnect failed (communication_channel: %s) %s",
                    self.communication_channel,
                    error_msg,
                )
        self.communication_channel = None

    async def request_description(self) -> GatewayDescriptor | None:
        """Request description from tunneling server."""
        description = DescriptionQuery(
            transport=self.transport,
            local_hpai=self.local_hpai,
        )
        await description.start()
        return description.gateway_descriptor

    @asynccontextmanager
    async def _send_ready(self) -> AsyncGenerator[None, None]:
        """Lock for sending frames sequentially and waiting for reconnect if applicable."""
        async with self._send_lock:
            # don't drop frames when reconnecting - wait for reconnect to finish
            if self._reconnect_task is not None:
                try:
                    await self._reconnect_task
                except asyncio.CancelledError:
                    pass
            yield

    async def send_cemi(self, cemi: CEMIFrame) -> None:
        """
        Send CEMI Frame to tunnelling server.

        A transport layer confirmation shall be awaited before sending the next telegram.
        """
        raw_cemi = cemi.to_knx()
        async with self._send_ready():
            try:
                await self._tunnelling_request(raw_cemi)
            finally:
                self._increase_sequence_number()

    async def _tunnelling_request(self, raw_cemi: bytes) -> None:
        """Send CEMI Frame to tunnelling server."""
        if self.communication_channel is None:
            raise CommunicationError(
                "Sending telegram failed. No active communication channel."
            )
        tunnelling_request = TunnellingRequest(
            communication_channel_id=self.communication_channel,
            sequence_counter=self.sequence_number,
            raw_cemi=raw_cemi,
        )
        await self._send_tunnelling_request(tunnelling_request)

    @abstractmethod
    async def _send_tunnelling_request(self, frame: TunnellingRequest) -> None:
        """Send TunnellingRequest frame to tunnelling device."""

    def _increase_sequence_number(self) -> None:
        """Increase sequence number."""
        self.sequence_number = self.sequence_number + 1 & 0xFF

    ####################
    #
    # INCOMING REQUESTS
    #
    ####################

    def _request_received(
        self, knxipframe: KNXIPFrame, source: HPAI, _transport: KNXIPTransport
    ) -> None:
        """Handle incoming requests."""
        if isinstance(knxipframe.body, TunnellingRequest):
            self._tunnelling_request_received(knxipframe.body)
        elif isinstance(knxipframe.body, DisconnectRequest):
            self._disconnect_request_received(knxipframe.body)
        else:
            logger.warning("Service not implemented: %s", knxipframe)

    def _tunnelling_request_received(
        self, tunneling_request: TunnellingRequest
    ) -> None:
        """Handle incoming tunnel request."""
        self.cemi_received_callback(tunneling_request.raw_cemi)

    def _disconnect_request_received(
        self, disconnect_request: DisconnectRequest
    ) -> None:
        """Handle incoming disconnect request."""
        logger.warning("Received DisconnectRequest from tunnelling server.")
        # We should not receive DisconnectRequest for other communication_channels
        # If we do we close our communication_channel before reconnection.
        if disconnect_request.communication_channel_id == self.communication_channel:
            disconnect_response = DisconnectResponse(
                communication_channel_id=self.communication_channel,
            )
            self.transport.send(KNXIPFrame.init_from_body(disconnect_response))
            self.communication_channel = None
        self._tunnel_lost()

    ####################
    #
    # HEARTBEAT
    #
    ####################

    def start_heartbeat(self) -> None:
        """Start heartbeat for monitoring state of tunnel, as suggested by 03.08.02 KNX Core 5.4."""
        self._heartbeat_task = asyncio.create_task(self.do_heartbeat())

    def stop_heartbeat(self) -> None:
        """Stop heartbeat task if running."""
        if self._heartbeat_task is not None:
            self._heartbeat_task.cancel()
            self._heartbeat_task = None

    async def do_heartbeat(self) -> None:
        """Heartbeat: Worker task, endless loop for sending heartbeat requests."""
        while True:
            try:
                await asyncio.sleep(HEARTBEAT_RATE)
                success, _ = await self._connectionstate_request()
                if not success:
                    await self._do_heartbeat_failed()
            except CommunicationError as err:
                logger.warning(err)
                self._tunnel_lost()

    async def _do_heartbeat_failed(self) -> None:
        """Heartbeat: handling error."""
        # first heartbeat failed - try 3 more times before disconnecting.
        for _heartbeats_failed in range(3):
            success, status = await self._connectionstate_request()
            if success:
                return
        # 3 retries failed
        if status is None:
            raise CommunicationError(
                "Heartbeat failed. No answer from tunnelling server."
            )
        raise CommunicationError(
            f"Heartbeat failed. Tunnelling server repeatedly responded with status: {status}"
        )


class UDPTunnel(_Tunnel):
    """Class for handling KNX/IP UDP tunnels."""

    __slots__ = (
        "_invalid_sequence_number_reconnect_task",
        "expected_sequence_number",
        "gateway_ip",
        "gateway_port",
        "local_ip",
        "local_port",
        "route_back",
    )

    connection_type = XknxConnectionType.TUNNEL_UDP
    transport: UDPTransport

    def __init__(
        self,
        xknx: XKNX,
        cemi_received_callback: CEMIBytesCallbackType,
        gateway_ip: str,
        gateway_port: int,
        local_ip: str,
        local_port: int = 0,
        route_back: bool = False,
        auto_reconnect: bool = True,
        auto_reconnect_wait: int = 3,
    ) -> None:
        """Initialize Tunnel class."""
        self.gateway_ip = gateway_ip
        self.gateway_port = gateway_port
        self.local_ip = local_ip
        self.local_port = local_port
        self.route_back = route_back
        super().__init__(
            xknx=xknx,
            cemi_received_callback=cemi_received_callback,
            auto_reconnect=auto_reconnect,
            auto_reconnect_wait=auto_reconnect_wait,
        )
        self.expected_sequence_number = 0
        self._invalid_sequence_number_reconnect_task: asyncio.Task[None] | None = None

    def _init_transport(self) -> None:
        """Initialize transport transport."""
        self.transport = UDPTransport(
            local_addr=(self.local_ip, self.local_port),
            remote_addr=(self.gateway_ip, self.gateway_port),
            multicast=False,
        )

    async def setup_tunnel(self) -> None:
        """Set up tunnel before sending a ConnectionRequest."""
        self.expected_sequence_number = 0

        if self.route_back:
            self.local_hpai = HPAI()
            return
        (local_addr, local_port) = self.transport.getsockname()
        self.local_hpai = HPAI(ip_addr=local_addr, port=local_port)

    def _prepare_disconnect(self) -> None:
        """Stop tunnel related tasks."""
        self._cancel_invalid_sequence_number_reconnect_schedule()
        super()._prepare_disconnect()

    ####################
    #
    # OUTGOING REQUESTS
    #
    ####################

    async def send_cemi(self, cemi: CEMIFrame) -> None:
        """
        Send CEMI Frame to tunnelling server - with retry mechanism for UDP connection.

        A transport layer confirmation shall be awaited before sending the next telegram.

        If a TUNNELLING_REQUEST frame is not confirmed within the TUNNELLING_REQUEST_TIMEOUT
        time of one (1) second then the frame shall be repeated once with the same sequence counter
        value by the sending KNXnet/IP device.

        If the KNXnet/IP device does not receive a TUNNELLING_ACK frame within the
        TUNNELLING_REQUEST_TIMEOUT (= 1 second) or the status of a received
        TUNNELLING_ACK frame signals any kind of error condition, the sending device
        shall repeat the TUNNELLING_REQUEST frame once and then terminate the
        connection by sending a DISCONNECT_REQUEST frame to the other devices
        control endpoint.
        """
        raw_cemi = cemi.to_knx()
        async with self._send_ready():
            try:
                try:
                    await self._tunnelling_request(raw_cemi)
                except TunnellingAckError as err:
                    logger.debug("First TunnellingRequest try failed. %s", err)
                else:
                    return

                try:
                    await self._tunnelling_request(raw_cemi)
                except TunnellingAckError as err:
                    logger.debug("Second TunnellingRequest try failed. %s", err)
                else:
                    return

                if self._reconnect_task is None:
                    self._tunnel_lost()
                if self._reconnect_task is None:
                    # _tunnel_lost() sets self._reconnect_task when auto-reconnect is True
                    raise CommunicationError(
                        "Sending TunnellingRequest failed twice. No reconnect.", True
                    )
                try:
                    await self._reconnect_task
                except asyncio.CancelledError:
                    raise CommunicationError(
                        "Sending TunnellingRequest failed twice. Reconnect was cancelled.",
                        True,
                    ) from None

                try:
                    await self._tunnelling_request(raw_cemi)
                except TunnellingAckError as err:
                    raise CommunicationError(
                        f"Third TunnellingRequest try failed after reconnect. {err}",
                        True,
                    ) from None

            finally:
                self._increase_sequence_number()

    async def _send_tunnelling_request(self, frame: TunnellingRequest) -> None:
        """Send Telegram to tunnelling device."""
        tunnelling = Tunnelling(
            transport=self.transport,
            data_endpoint=self._data_endpoint_addr,
            tunnelling_request=frame,
        )
        await tunnelling.start()

        if not tunnelling.success:
            if error_code := tunnelling.response_status_code:
                reason = (
                    f"Received TUNNELLING_ACK with error code: {error_code.name} "
                    f"for frame {frame}"
                )
            else:
                reason = (
                    "Did not receive a TUNNELLING_ACK within 1 second for "
                    f"frame with sequence_counter={frame.sequence_counter}"
                )
            raise TunnellingAckError(reason)

    ####################
    #
    # INCOMING REQUESTS
    #
    ####################

    def _invalid_sequence_number_reconnect_schedule(
        self,
        seconds: int = 2,  # 2x ACK timeout
    ) -> None:
        """
        Schedule a reconnect or disconnect when a TunnellingRequest with invalid sequence number was received.

        To be canceled if a valid frame is received before the timeout.
        """

        async def _schedule_tunnel_lost() -> None:
            await asyncio.sleep(seconds)
            logger.warning(
                "TunnellingRequest with expected sequence number %s was not received within %s seconds "
                "after an unexpected sequence number was received. Closing tunnel connection.",
                self.expected_sequence_number,
                seconds,
            )
            # Clear the stored task reference before calling _tunnel_lost() so
            # that the reconnect logic does not cancel this running task itself.
            self._invalid_sequence_number_reconnect_task = None
            self._tunnel_lost()

        if (
            self._invalid_sequence_number_reconnect_task is None
            and self._reconnect_task is None
        ):
            self._invalid_sequence_number_reconnect_task = asyncio.create_task(
                _schedule_tunnel_lost()
            )

    def _cancel_invalid_sequence_number_reconnect_schedule(self) -> None:
        """Cancel scheduled reconnect or disconnect."""
        if self._invalid_sequence_number_reconnect_task is not None:
            self._invalid_sequence_number_reconnect_task.cancel()
            self._invalid_sequence_number_reconnect_task = None
            logger.debug("Canceled scheduled reconnect for invalid sequence number.")

    def _tunnelling_request_received(
        self, tunneling_request: TunnellingRequest
    ) -> None:
        """Handle incoming tunnelling request."""
        if tunneling_request.sequence_counter == self.expected_sequence_number:
            self.expected_sequence_number = self.expected_sequence_number + 1 & 0xFF
            # valid frame received - cancel scheduled reconnect for invalid sequence number
            self._cancel_invalid_sequence_number_reconnect_schedule()

            self._send_tunnelling_ack(
                tunneling_request.communication_channel_id,
                tunneling_request.sequence_counter,
            )
            super()._tunnelling_request_received(tunneling_request)
            return
        if (
            tunneling_request.sequence_counter
            == self.expected_sequence_number - 1 & 0xFF
        ):
            self._send_tunnelling_ack(
                tunneling_request.communication_channel_id,
                tunneling_request.sequence_counter,
            )
            logger.debug(
                "Received TunnellingRequest with sequence number one less than expected. "
                "Discarding frame: %s",
                tunneling_request,
            )
            return
        logger.warning(
            "Received TunnellingRequest with sequence number not equal to expected: %s. "
            "Discarding frame: %s",
            self.expected_sequence_number,
            tunneling_request,
        )
        # Tunnelling server should repeat that frame and disconnect after that was also not ACKed.
        # According to KNX specification we should just drop the frame.
        # Some interfaces don't initiate a disconnect here so we take a shortcut and disconnect
        # ourselves when we don't receive the expected frame after a short delay (mixed up frames?).
        self._invalid_sequence_number_reconnect_schedule()

    def _send_tunnelling_ack(
        self, communication_channel_id: int, sequence_counter: int
    ) -> None:
        """Send tunnelling ACK after tunnelling request received."""
        ack = TunnellingAck(
            communication_channel_id=communication_channel_id,
            sequence_counter=sequence_counter,
        )
        self.transport.send(
            KNXIPFrame.init_from_body(ack), addr=self._data_endpoint_addr
        )


class TCPTunnel(_Tunnel):
    """Class for handling KNX/IP TCP tunnels."""

    __slots__ = ("gateway_ip", "gateway_port")

    connection_type = XknxConnectionType.TUNNEL_TCP
    transport: TCPTransport

    def __init__(
        self,
        xknx: XKNX,
        cemi_received_callback: CEMIBytesCallbackType,
        gateway_ip: str,
        gateway_port: int,
        individual_address: IndividualAddress | None = None,
        auto_reconnect: bool = True,
        auto_reconnect_wait: int = 3,
    ) -> None:
        """Initialize Tunnel class."""
        self.gateway_ip = gateway_ip
        self.gateway_port = gateway_port
        super().__init__(
            xknx=xknx,
            cemi_received_callback=cemi_received_callback,
            auto_reconnect=auto_reconnect,
            auto_reconnect_wait=auto_reconnect_wait,
        )
        # TCP always uses 0.0.0.0:0
        self.local_hpai = HPAI(protocol=HostProtocol.IPV4_TCP)
        self._requested_address = individual_address

    def _init_transport(self) -> None:
        """Initialize transport transport."""
        self.transport = TCPTransport(
            remote_addr=(self.gateway_ip, self.gateway_port),
            connection_lost_cb=self._tunnel_lost,
        )

    async def setup_tunnel(self) -> None:
        """Set up tunnel before sending a ConnectionRequest."""

    async def _send_tunnelling_request(self, frame: TunnellingRequest) -> None:
        """Send Telegram to tunnelling device."""
        self.transport.send(KNXIPFrame.init_from_body(frame))


class SecureTunnel(TCPTunnel):
    """Class for handling KNX/IP secure TCP tunnels."""

    __slots__ = ("_device_authentication_password", "_user_id", "_user_password")

    connection_type = XknxConnectionType.TUNNEL_SECURE
    transport: SecureSession

    def __init__(
        self,
        xknx: XKNX,
        cemi_received_callback: CEMIBytesCallbackType,
        gateway_ip: str,
        gateway_port: int,
        user_id: int,
        user_password: str,
        auto_reconnect: bool = True,
        auto_reconnect_wait: int = 3,
        device_authentication_password: str | None = None,
    ) -> None:
        """Initialize SecureTunnel class."""
        self._device_authentication_password = device_authentication_password
        self._user_id = user_id
        self._user_password = user_password
        super().__init__(
            xknx=xknx,
            cemi_received_callback=cemi_received_callback,
            gateway_ip=gateway_ip,
            gateway_port=gateway_port,
            auto_reconnect=auto_reconnect,
            auto_reconnect_wait=auto_reconnect_wait,
        )

    def _init_transport(self) -> None:
        """Initialize transport transport."""
        self.transport = SecureSession(
            remote_addr=(self.gateway_ip, self.gateway_port),
            user_id=self._user_id,
            user_password=self._user_password,
            device_authentication_password=self._device_authentication_password,
            connection_lost_cb=self._tunnel_lost,
        )



"""
GatewayScanner is an abstraction for searching for KNX/IP devices on the local network.

It walks through all network interfaces and sends UDP multicast
SearchRequest and SearchRequestExtended frames.
"""

from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator
from functools import partial
import logging
from typing import TYPE_CHECKING

from xknx.exceptions import XKNXException
from xknx.io import util
from xknx.knxip import (
    HPAI,
    SRP,
    DIBServiceFamily,
    DIBTypeCode,
    KNXIPFrame,
    KNXIPServiceType,
    SearchRequest,
    SearchRequestExtended,
    SearchResponse,
    SearchResponseExtended,
)
from xknx.knxip.dib import (
    DIB,
    DIBDeviceInformation,
    DIBSecuredServiceFamilies,
    DIBSuppSVCFamilies,
    DIBTunnelingInfo,
    TunnelingSlotStatus,
)
from xknx.telegram import IndividualAddress
from xknx.util import asyncio_timeout

from .transport import UDPTransport

if TYPE_CHECKING:
    from xknx.xknx import XKNX

logger = logging.getLogger("xknx.log")


class GatewayDescriptor:
    """Used to return information about the discovered gateways."""

    def __init__(
        self,
        ip_addr: str,
        port: int,
        local_ip: str = "",
        local_interface: str = "",
        name: str = "UNKNOWN",
        supports_routing: bool = False,
        supports_tunnelling: bool = False,
        supports_tunnelling_tcp: bool = False,
        supports_secure: bool = False,
        individual_address: IndividualAddress | None = None,
    ) -> None:
        """Initialize GatewayDescriptor class."""
        self.name = name
        self.ip_addr = ip_addr
        self.port = port
        self.individual_address = individual_address
        self.local_interface = local_interface
        self.local_ip = local_ip
        self.supports_routing = supports_routing
        self.supports_tunnelling = supports_tunnelling
        self.supports_tunnelling_tcp = supports_tunnelling_tcp
        self.supports_secure = supports_secure

        self.core_version: int = 0
        self.routing_requires_secure: bool | None = None
        self.tunnelling_requires_secure: bool | None = None
        self.tunnelling_slots: dict[IndividualAddress, TunnelingSlotStatus] = {}

    def parse_dibs(self, dibs: list[DIB]) -> None:
        """Parse DIBs for gateway information."""
        for dib in dibs:
            if isinstance(dib, DIBDeviceInformation):
                self.name = dib.name
                self.individual_address = dib.individual_address
                continue
            if isinstance(dib, DIBSuppSVCFamilies):
                self.core_version = dib.version(DIBServiceFamily.CORE) or 0
                self.supports_routing = dib.supports(DIBServiceFamily.ROUTING)
                if _tunnelling_version := dib.version(DIBServiceFamily.TUNNELING):
                    self.supports_tunnelling = True
                    self.supports_tunnelling_tcp = _tunnelling_version >= 2
                self.supports_secure = dib.supports(
                    DIBServiceFamily.SECURITY, version=1
                )
                continue
            if isinstance(dib, DIBSecuredServiceFamilies):
                self.tunnelling_requires_secure = dib.supports(
                    DIBServiceFamily.TUNNELING
                )
                self.routing_requires_secure = dib.supports(DIBServiceFamily.ROUTING)
                continue
            if isinstance(dib, DIBTunnelingInfo):
                self.tunnelling_slots = dib.slots
                continue

    def __repr__(self) -> str:
        """Return object as representation string."""
        return (
            "GatewayDescriptor(\n"
            f"    name={self.name},\n"
            f"    ip_addr={self.ip_addr},\n"
            f"    port={self.port},\n"
            f"    individual_address={self.individual_address}\n"
            f"    local_interface={self.local_interface},\n"
            f"    local_ip={self.local_ip},\n"
            f"    core_version={self.core_version},\n"
            f"    supports_routing={self.supports_routing},\n"
            f"    supports_tunnelling={self.supports_tunnelling},\n"
            f"    supports_tunnelling_tcp={self.supports_tunnelling_tcp},\n"
            f"    supports_secure={self.supports_secure},\n"
            f"    routing_requires_secure={self.routing_requires_secure}\n"
            f"    tunnelling_requires_secure={self.tunnelling_requires_secure}\n"
            f"    tunnelling_slots={self.tunnelling_slots}\n"
            ")"
        )

    def __str__(self) -> str:
        """Return object as readable string."""
        return f"{self.individual_address} - {self.name} @ {self.ip_addr}:{self.port}"


class GatewayScanFilter:
    """
    Filter to limit gateway scan results.

    If `name` doesn't match the gateway name, the gateway will be ignored.

    Connection methods are treated as OR if `True` is set for multiple methods.
    Non-secure methods don't match if secure is required.
    """

    def __init__(
        self,
        name: str | None = None,
        tunnelling: bool | None = True,
        tunnelling_tcp: bool | None = True,
        routing: bool | None = True,
        secure_tunnelling: bool | None = True,
        secure_routing: bool | None = True,
    ) -> None:
        """Initialize GatewayScanFilter class."""
        self.name = name
        self.tunnelling = tunnelling
        self.tunnelling_tcp = tunnelling_tcp
        self.routing = routing
        self.secure_tunnelling = secure_tunnelling
        self.secure_routing = secure_routing

    def match(self, gateway: GatewayDescriptor) -> bool:
        """Check whether the device is a gateway and given GatewayDescriptor matches the filter."""
        if self.name is not None and self.name != gateway.name:
            return False
        return (
            bool(
                self.tunnelling
                and gateway.supports_tunnelling
                and not gateway.tunnelling_requires_secure
            )
            or bool(
                self.tunnelling_tcp
                and gateway.supports_tunnelling_tcp
                and not gateway.tunnelling_requires_secure
            )
            or bool(
                self.routing
                and gateway.supports_routing
                and not gateway.routing_requires_secure
            )
            or bool(
                self.secure_tunnelling
                and gateway.supports_tunnelling_tcp
                and gateway.tunnelling_requires_secure
            )
            or bool(
                self.secure_routing
                and gateway.supports_routing
                and gateway.routing_requires_secure
            )
        )

    def __eq__(self, other: object) -> bool:
        """Equality for GatewayScanFilter class."""
        return self.__dict__ == other.__dict__


class GatewayScanner:
    """Class for searching KNX/IP devices."""

    def __init__(
        self,
        xknx: XKNX,
        local_ip: str | None = None,
        timeout_in_seconds: float = 3.0,
        stop_on_found: int | None = None,
        scan_filter: GatewayScanFilter | None = None,
    ) -> None:
        """Initialize GatewayScanner class."""
        self.xknx = xknx
        self.local_ip = local_ip
        self.timeout_in_seconds = timeout_in_seconds
        self.stop_on_found = stop_on_found
        self.scan_filter = scan_filter or GatewayScanFilter()
        self.found_gateways: dict[HPAI, GatewayDescriptor] = {}
        self._response_received_event = asyncio.Event()

    async def scan(self) -> list[GatewayDescriptor]:
        """Scan and return a list of GatewayDescriptors on success."""
        await self._scan()
        return list(self.found_gateways.values())

    async def async_scan(self) -> AsyncGenerator[GatewayDescriptor, None]:
        """Search and yield found gateways."""
        queue: asyncio.Queue[GatewayDescriptor | None] = asyncio.Queue()
        scan_task = asyncio.create_task(self._scan(queue=queue))
        try:
            while True:
                gateway = await queue.get()
                if gateway is None:
                    return
                yield gateway
        finally:
            # cleanup after GeneratorExit or XKNXExceptions
            if not scan_task.done():
                scan_task.cancel()
            await scan_task  # to bubble up exceptions

    async def _scan(
        self, queue: asyncio.Queue[GatewayDescriptor | None] | None = None
    ) -> None:
        """Scan for gateways."""
        _local_ip = self.local_ip or await util.get_default_local_ip(
            remote_ip=self.xknx.multicast_group
        )
        if _local_ip is None:
            if queue is not None:
                queue.put_nowait(None)
            raise XKNXException("No usable network interface found.")
        local_ip = await util.validate_ip(_local_ip)
        interface_name = util.get_local_interface_name(local_ip=local_ip)
        logger.debug("Searching on %s / %s", interface_name, local_ip)

        udp_transport = UDPTransport(
            local_addr=(local_ip, 0),
            remote_addr=(self.xknx.multicast_group, self.xknx.multicast_port),
        )
        udp_transport.register_callback(
            partial(self._response_rec_callback, interface=interface_name, queue=queue),
            [
                KNXIPServiceType.SEARCH_RESPONSE,
                KNXIPServiceType.SEARCH_RESPONSE_EXTENDED,
            ],
        )
        try:
            await self._send_search_requests(udp_transport=udp_transport)
            async with asyncio_timeout(self.timeout_in_seconds):
                await self._response_received_event.wait()
        except asyncio.TimeoutError:
            pass
        except asyncio.CancelledError:
            pass
        finally:
            udp_transport.stop()
            if queue is not None:
                queue.put_nowait(None)

    @staticmethod
    async def _send_search_requests(udp_transport: UDPTransport) -> None:
        """Send search requests on a specific interface."""
        await udp_transport.connect()
        discovery_endpoint = HPAI(*udp_transport.getsockname())
        # send SearchRequestExtended requesting needed DIBs
        search_request_extended = SearchRequestExtended(
            discovery_endpoint=discovery_endpoint,
            srps=[
                SRP.request_device_description(
                    [
                        DIBTypeCode.DEVICE_INFO,
                        DIBTypeCode.SUPP_SVC_FAMILIES,
                        DIBTypeCode.SECURED_SERVICE_FAMILIES,
                        DIBTypeCode.TUNNELING_INFO,
                    ]
                )
            ],
        )
        udp_transport.send(KNXIPFrame.init_from_body(search_request_extended))
        # send SearchRequest for Core-V1 devices
        search_request = SearchRequest(discovery_endpoint=discovery_endpoint)
        udp_transport.send(KNXIPFrame.init_from_body(search_request))

    def _response_rec_callback(
        self,
        knx_ip_frame: KNXIPFrame,
        source: HPAI,
        udp_transport: UDPTransport,
        interface: str = "",
        queue: asyncio.Queue[GatewayDescriptor | None] | None = None,
    ) -> None:
        """Verify and handle knxipframe. Callback from internal udp_transport."""
        if not isinstance(knx_ip_frame.body, SearchResponse | SearchResponseExtended):
            logger.warning("Could not understand knxipframe")
            return

        # skip non-extended SearchResponse for Core-V2 devices
        if knx_ip_frame.header.service_type_ident == KNXIPServiceType.SEARCH_RESPONSE:
            if svc_families_dib := next(
                (
                    dib
                    for dib in knx_ip_frame.body.dibs
                    if isinstance(dib, DIBSuppSVCFamilies)
                ),
                None,
            ):
                if svc_families_dib.supports(DIBServiceFamily.CORE, version=2):
                    logger.debug("Skipping SearchResponse for Core-V2 device")
                    return

        gateway = GatewayDescriptor(
            ip_addr=knx_ip_frame.body.control_endpoint.ip_addr,
            port=knx_ip_frame.body.control_endpoint.port,
            local_ip=udp_transport.local_addr[0],
            local_interface=interface,
        )
        gateway.parse_dibs(knx_ip_frame.body.dibs)

        logger.debug("Found KNX/IP device at %s: %s", source, repr(gateway))
        if self.scan_filter.match(gateway):
            self.found_gateways[knx_ip_frame.body.control_endpoint] = gateway
            if queue is not None:
                queue.put_nowait(gateway)
            if self.stop_on_found and len(self.found_gateways) >= self.stop_on_found:
                self._response_received_event.set()
